# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core.operand import OperandStage, MapReduceOperand
from ...utils import lazy_import, calc_nsplits
from ...serialization.serializables import (
    AnyField,
    Int32Field,
    ListField,
    StringField,
    BoolField,
)
from ...tensor.base.psrs import PSRSOperandMixin
from ..core import IndexValue, OutputType
from ..utils import standardize_range_index, parse_index, is_cudf
from ..operands import DataFrameOperandMixin, DataFrameOperand, DataFrameShuffleProxy


cudf = lazy_import("cudf")

_PSRS_DISTINCT_COL = "__PSRS_TMP_DISTINCT_COL"


class _Largest:
    """
    This util class resolve TypeError when
    comparing strings with None values
    """

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return self is not other


_largest = _Largest()


class _ReversedValue:
    def __init__(self, value):
        self._value = value

    def __lt__(self, other):
        if type(other) is _ReversedValue:
            # may happen when call searchsorted
            return self._value >= other._value
        return self._value >= other

    def __gt__(self, other):
        return self._value <= other

    def __repr__(self):
        return repr(self._value)


class DataFramePSRSOperandMixin(DataFrameOperandMixin, PSRSOperandMixin):
    @classmethod
    def _collect_op_properties(cls, op):
        from .sort_values import DataFrameSortValues

        if isinstance(op, DataFrameSortValues):
            properties = dict(
                sort_type="sort_values",
                axis=op.axis,
                by=op.by,
                ascending=op.ascending,
                inplace=op.inplace,
                na_position=op.na_position,
                gpu=op.is_gpu(),
            )
        else:
            properties = dict(
                sort_type="sort_index",
                axis=op.axis,
                level=op.level,
                ascending=op.ascending,
                inplace=op.inplace,
                na_position=op.na_position,
                sort_remaining=op.sort_remaining,
                gpu=op.is_gpu(),
            )
        return properties

    @classmethod
    def local_sort_and_regular_sample(
        cls, op, in_data, axis_chunk_shape, axis_offsets, out_idx
    ):
        # stage 1: local sort and regular samples collected
        sorted_chunks, indices_chunks, sampled_chunks = [], [], []
        for i in range(axis_chunk_shape):
            in_chunk = in_data.chunks[i]
            kind = None if op.psrs_kinds is None else op.psrs_kinds[0]
            chunk_op = DataFramePSRSSortRegularSample(
                kind=kind,
                n_partition=axis_chunk_shape,
                output_types=op.output_types,
                **cls._collect_op_properties(op)
            )
            kws = []
            sort_shape = in_chunk.shape
            kws.append(
                {
                    "shape": sort_shape,
                    "index_value": in_chunk.index_value,
                    "index": in_chunk.index,
                }
            )
            if chunk_op.sort_type == "sort_values":
                sampled_shape = (
                    (axis_chunk_shape, len(op.by)) if op.by else (axis_chunk_shape,)
                )
            else:
                sampled_shape = (
                    (axis_chunk_shape, sort_shape[1])
                    if len(sort_shape) == 2
                    else (axis_chunk_shape,)
                )
            kws.append(
                {
                    "shape": sampled_shape,
                    "index_value": in_chunk.index_value,
                    "index": (i,),
                    "type": "regular_sampled",
                }
            )
            if op.outputs[0].ndim == 2:
                kws[0].update(
                    {"columns_value": in_chunk.columns_value, "dtypes": in_chunk.dtypes}
                )
                kws[1].update(
                    {"columns_value": in_chunk.columns_value, "dtypes": in_chunk.dtypes}
                )
            else:
                kws[0].update(({"dtype": in_chunk.dtype, "name": in_chunk.name}))
                kws[1].update({"dtype": in_chunk.dtype})

            chunks = chunk_op.new_chunks([in_chunk], kws=kws, output_limit=len(kws))
            sort_chunk, sampled_chunk = chunks
            sorted_chunks.append(sort_chunk)
            sampled_chunks.append(sampled_chunk)
        return sorted_chunks, indices_chunks, sampled_chunks

    @classmethod
    def concat_and_pivot(
        cls, op, axis_chunk_shape, out_idx, sorted_chunks, sampled_chunks
    ):
        from .sort_values import DataFrameSortValues

        # stage 2: gather and merge samples, choose and broadcast p-1 pivots
        kind = None if op.psrs_kinds is None else op.psrs_kinds[1]
        if isinstance(op, DataFrameSortValues):
            output_types = op.output_types
        else:
            output_types = [OutputType.index]
        concat_pivot_op = DataFramePSRSConcatPivot(
            kind=kind,
            n_partition=axis_chunk_shape,
            output_types=output_types,
            **cls._collect_op_properties(op)
        )
        concat_pivot_shape = (
            sorted_chunks[0].shape[: op.axis]
            + (axis_chunk_shape - 1,)
            + sorted_chunks[0].shape[op.axis + 1 :]
        )
        concat_pivot_index = out_idx[: op.axis] + (0,) + out_idx[op.axis :]
        concat_pivot_chunk = concat_pivot_op.new_chunk(
            sampled_chunks,
            shape=concat_pivot_shape,
            index=concat_pivot_index,
        )
        return concat_pivot_chunk

    @classmethod
    def partition_local_data(
        cls, op, axis_chunk_shape, sorted_chunks, indices_chunks, concat_pivot_chunk
    ):
        # stage 3: Local data is partitioned
        partition_chunks = []
        length = len(sorted_chunks)
        for i in range(length):
            chunk_inputs = [sorted_chunks[i], concat_pivot_chunk]
            partition_shuffle_map = DataFramePSRSShuffle(
                n_partition=axis_chunk_shape,
                stage=OperandStage.map,
                output_types=op.output_types,
                **cls._collect_op_properties(op)
            )
            if isinstance(chunk_inputs[0].index_value.value, IndexValue.RangeIndex):
                index_value = parse_index(pd.Index([], dtype=np.int64))
            else:
                index_value = chunk_inputs[0].index_value
            kw = dict(
                shape=chunk_inputs[0].shape,
                index=chunk_inputs[0].index,
                index_value=index_value,
            )
            if op.outputs[0].ndim == 2:
                kw.update(
                    dict(
                        columns_value=chunk_inputs[0].columns_value,
                        dtypes=chunk_inputs[0].dtypes,
                    )
                )
            else:
                kw.update(dict(dtype=chunk_inputs[0].dtype, name=chunk_inputs[0].name))
            partition_chunk = partition_shuffle_map.new_chunk(chunk_inputs, **kw)
            partition_chunks.append(partition_chunk)
        return partition_chunks

    @classmethod
    def partition_merge_data(
        cls, op, need_align, return_value, partition_chunks, proxy_chunk
    ):
        # stage 4: all *ith* classes are gathered and merged
        partition_sort_chunks, partition_indices_chunks, sort_info_chunks = [], [], []
        for i, partition_chunk in enumerate(partition_chunks):
            kind = None if op.psrs_kinds is None else op.psrs_kinds[2]
            partition_shuffle_reduce = DataFramePSRSShuffle(
                stage=OperandStage.reduce,
                kind=kind,
                reducer_index=(i,),
                reducer_ordinal=i,
                n_reducers=len(partition_chunks),
                output_types=op.output_types,
                **cls._collect_op_properties(op)
            )
            chunk_shape = list(partition_chunk.shape)
            chunk_shape[op.axis] = np.nan

            kw = dict(
                shape=tuple(chunk_shape),
                index=partition_chunk.index,
                index_value=partition_chunk.index_value,
            )
            if op.outputs[0].ndim == 2:
                kw.update(
                    dict(
                        columns_value=partition_chunk.columns_value,
                        dtypes=partition_chunk.dtypes,
                    )
                )
            else:
                kw.update(dict(dtype=partition_chunk.dtype, name=partition_chunk.name))
            cs = partition_shuffle_reduce.new_chunks([proxy_chunk], **kw)

            partition_sort_chunks.append(cs[0])
        return partition_sort_chunks, partition_indices_chunks, sort_info_chunks

    @classmethod
    def _tile_psrs(cls, op, in_data):
        out = op.outputs[0]
        in_df, axis_chunk_shape, _, _ = yield from cls.preprocess(op, in_data=in_data)

        # stage 1: local sort and regular samples collected
        sorted_chunks, _, sampled_chunks = cls.local_sort_and_regular_sample(
            op, in_df, axis_chunk_shape, None, None
        )

        # stage 2: gather and merge samples, choose and broadcast p-1 pivots
        concat_pivot_chunk = cls.concat_and_pivot(
            op,
            axis_chunk_shape,
            (0,) if in_df.ndim == 2 else (),
            sorted_chunks,
            sampled_chunks,
        )

        # stage 3: Local data is partitioned
        partition_chunks = cls.partition_local_data(
            op, axis_chunk_shape, sorted_chunks, None, concat_pivot_chunk
        )

        proxy_chunk = DataFrameShuffleProxy(output_types=op.output_types).new_chunk(
            partition_chunks, shape=()
        )

        # stage 4: all *ith* classes are gathered and merged
        partition_sort_chunks = cls.partition_merge_data(
            op, False, None, partition_chunks, proxy_chunk
        )[0]

        if op.ignore_index:
            yield partition_sort_chunks
            chunks = standardize_range_index(partition_sort_chunks, axis=op.axis)
        else:
            chunks = partition_sort_chunks

        nsplits = calc_nsplits({c.index: c.shape for c in chunks})
        if op.outputs[0].ndim == 2:
            new_op = op.copy()
            return new_op.new_dataframes(
                op.inputs,
                shape=out.shape,
                chunks=chunks,
                nsplits=nsplits,
                index_value=out.index_value,
                columns_value=out.columns_value,
                dtypes=out.dtypes,
            )
        else:
            new_op = op.copy()
            return new_op.new_seriess(
                op.inputs,
                shape=out.shape,
                chunks=chunks,
                nsplits=nsplits,
                index_value=out.index_value,
                dtype=out.dtype,
                name=out.name,
            )


def execute_sort_values(data, op, inplace=None, by=None):
    if inplace is None:
        inplace = op.inplace
    # ignore_index is new in Pandas version 1.0.0.
    ignore_index = getattr(op, "ignore_index", False)
    if isinstance(data, (pd.DataFrame, pd.Series)):
        kwargs = dict(
            axis=op.axis,
            ascending=op.ascending,
            ignore_index=ignore_index,
            na_position=op.na_position,
            kind=op.kind,
        )
        if isinstance(data, pd.DataFrame):
            kwargs["by"] = by if by is not None else op.by
        if inplace:
            kwargs["inplace"] = True
            try:
                data.sort_values(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                data.sort_values(**kwargs)
            return data
        else:
            try:
                return data.sort_values(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                return data.sort_values(**kwargs)

    else:  # pragma: no cover
        # cudf doesn't support axis and kind
        if isinstance(data, cudf.DataFrame):
            return data.sort_values(
                op.by, ascending=op.ascending, na_position=op.na_position
            )
        else:
            return data.sort_values(ascending=op.ascending, na_position=op.na_position)


def execute_sort_index(data, op, inplace=None):
    if inplace is None:
        inplace = op.inplace
    # ignore_index is new in Pandas version 1.0.0.
    ignore_index = getattr(op, "ignore_index", False)
    if isinstance(data, (pd.DataFrame, pd.Series)):
        kwargs = dict(
            level=op.level,
            ascending=op.ascending,
            ignore_index=ignore_index,
            na_position=op.na_position,
            kind=op.kind,
            sort_remaining=op.sort_remaining,
        )
        if inplace:
            kwargs["inplace"] = True
            try:
                data.sort_index(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                data.sort_index(**kwargs)
            return data
        else:
            try:
                return data.sort_index(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                return data.sort_index(**kwargs)

    else:  # pragma: no cover
        # cudf only support ascending
        return data.sort_index(ascending=op.ascending)


class DataFramePSRSChunkOperand(DataFrameOperand):
    # sort type could be 'sort_values' or 'sort_index'
    sort_type = StringField("sort_type")

    axis = Int32Field("axis")
    by = ListField("by", default=None)
    ascending = AnyField("ascending")
    inplace = BoolField("inplace")
    kind = StringField("kind")
    na_position = StringField("na_position")

    # for sort_index
    level = ListField("level")
    sort_remaining = BoolField("sort_remaining")

    n_partition = Int32Field("n_partition")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)


class DataFramePSRSSortRegularSample(DataFramePSRSChunkOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_SORT_REGULAR_SMAPLE

    @property
    def output_limit(self):
        return 2

    @classmethod
    def execute(cls, ctx, op):
        a = ctx[op.inputs[0].key]
        xdf = pd if isinstance(a, (pd.DataFrame, pd.Series)) else cudf

        if len(a) == 0:
            # when chunk is empty, return the empty chunk itself
            ctx[op.outputs[0].key] = ctx[op.outputs[-1].key] = a
            return

        if op.sort_type == "sort_values":
            ctx[op.outputs[0].key] = res = execute_sort_values(a, op)
        else:
            ctx[op.outputs[0].key] = res = execute_sort_index(a, op)

        by = op.by
        add_distinct_col = bool(int(os.environ.get("PSRS_DISTINCT_COL", "0")))
        if (
            add_distinct_col
            and isinstance(a, xdf.DataFrame)
            and op.sort_type == "sort_values"
        ):
            # when running under distributed mode, we introduce an extra column
            # to make sure pivots are distinct
            chunk_idx = op.inputs[0].index[0]
            distinct_col = (
                _PSRS_DISTINCT_COL
                if a.columns.nlevels == 1
                else (_PSRS_DISTINCT_COL,) + ("",) * (a.columns.nlevels - 1)
            )
            res[distinct_col] = np.arange(
                chunk_idx << 32, (chunk_idx << 32) + len(a), dtype=np.int64
            )
            by = list(by) + [distinct_col]

        n = op.n_partition
        if op.sort_type == "sort_values" and a.shape[op.axis] < n:
            num = n // a.shape[op.axis] + 1
            res = execute_sort_values(xdf.concat([res] * num), op, by=by)

        w = res.shape[op.axis] * 1.0 / (n + 1)
        slc = np.linspace(
            max(w - 1, 0), res.shape[op.axis] - 1, num=n, endpoint=False
        ).astype(int)
        if op.axis == 1:
            slc = (slice(None), slc)
        if op.sort_type == "sort_values":
            # do regular sample
            if op.by is not None:
                ctx[op.outputs[-1].key] = res[by].iloc[slc]
            else:
                ctx[op.outputs[-1].key] = res.iloc[slc]
        else:
            # do regular sample
            ctx[op.outputs[-1].key] = res.iloc[slc]


class DataFramePSRSConcatPivot(DataFramePSRSChunkOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_CONCAT_PIVOT

    @property
    def output_limit(self):
        return 1

    @classmethod
    def execute(cls, ctx, op):
        inputs = [ctx[c.key] for c in op.inputs if len(ctx[c.key]) > 0]
        if len(inputs) == 0:
            # corner case: nothing sampled, we need to do nothing
            ctx[op.outputs[-1].key] = ctx[op.inputs[0].key]
            return

        xdf = pd if isinstance(inputs[0], (pd.DataFrame, pd.Series)) else cudf

        a = xdf.concat(inputs, axis=op.axis)
        p = len(inputs)
        assert a.shape[op.axis] == p * len(op.inputs)

        slc = np.linspace(
            p - 1, a.shape[op.axis] - 1, num=len(op.inputs) - 1, endpoint=False
        ).astype(int)
        if op.axis == 1:
            slc = (slice(None), slc)
        if op.sort_type == "sort_values":
            a = execute_sort_values(a, op, inplace=False)
            ctx[op.outputs[-1].key] = a.iloc[slc]
        else:
            a = execute_sort_index(a, op, inplace=False)
            ctx[op.outputs[-1].key] = a.index[slc]


class DataFramePSRSShuffle(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_SHUFFLE

    sort_type = StringField("sort_type")

    # for shuffle map
    axis = Int32Field("axis")
    by = ListField("by")
    ascending = AnyField("ascending")
    inplace = BoolField("inplace")
    na_position = StringField("na_position")
    n_partition = Int32Field("n_partition")

    # for sort_index
    level = ListField("level")
    sort_remaining = BoolField("sort_remaining")

    # for shuffle reduce
    kind = StringField("kind")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @property
    def output_limit(self):
        return 1

    @staticmethod
    def _calc_poses(src_cols, pivots, ascending=True):
        if isinstance(ascending, list):
            for asc, col in zip(ascending, pivots.columns):
                # Make pivots available to use ascending order when mixed order specified
                if not asc:
                    if pd.api.types.is_numeric_dtype(pivots.dtypes[col]):
                        # for numeric dtypes, convert to negative is more efficient
                        pivots[col] = -pivots[col]
                        src_cols[col] = -src_cols[col]
                    else:
                        # for other types, convert to ReversedValue
                        pivots[col] = pivots[col].map(
                            lambda x: x
                            if type(x) is _ReversedValue
                            else _ReversedValue(x)
                        )
            ascending = True

        records = src_cols.to_records(index=False)
        p_records = pivots.to_records(index=False)
        if ascending:
            poses = records.searchsorted(p_records, side="right")
        else:
            poses = len(records) - records[::-1].searchsorted(p_records, side="right")
        del records, p_records
        return poses

    @classmethod
    def _execute_dataframe_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if len(a) == 0:
            # when the chunk is empty, no slices can be produced
            for i in range(op.n_partition):
                ctx[out.key, (i,)] = a
            return

        # use numpy.searchsorted to find split positions.
        by = op.by

        distinct_col = (
            _PSRS_DISTINCT_COL
            if a.columns.nlevels == 1
            else (_PSRS_DISTINCT_COL,) + ("",) * (a.columns.nlevels - 1)
        )
        if distinct_col in a.columns:
            by = list(by) + [distinct_col]

        try:
            poses = cls._calc_poses(a[by], pivots, op.ascending)
        except TypeError:
            poses = cls._calc_poses(
                a[by].fillna(_largest), pivots.fillna(_largest), op.ascending
            )

        poses = (None,) + tuple(poses) + (None,)
        for i in range(op.n_partition):
            values = a.iloc[poses[i] : poses[i + 1]]
            if is_cudf(values):  # pragma: no cover
                values = values.copy()
            ctx[out.key, (i,)] = values

    @classmethod
    def _calc_series_poses(cls, s, pivots, ascending=True):
        if ascending:
            poses = s.searchsorted(pivots, side="right")
        else:
            poses = len(s) - s.iloc[::-1].searchsorted(pivots, side="right")
        return poses

    @classmethod
    def _execute_series_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if len(a) == 0:
            # when the chunk is empty, no slices can be produced
            for i in range(op.n_partition):
                ctx[out.key, (i,)] = a
            return

        if isinstance(a, pd.Series):
            try:
                poses = cls._calc_series_poses(a, pivots, ascending=op.ascending)
            except TypeError:
                filled_a = a.fillna(_largest)
                filled_pivots = pivots.fillna(_largest)
                poses = cls._calc_series_poses(
                    filled_a, filled_pivots, ascending=op.ascending
                )
            poses = (None,) + tuple(poses) + (None,)
            for i in range(op.n_partition):
                values = a.iloc[poses[i] : poses[i + 1]]
                ctx[out.key, (i,)] = values

    @classmethod
    def _execute_sort_index_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if op.ascending:
            poses = a.index.searchsorted(list(pivots), side="right")
        else:
            poses = len(a) - a.index[::-1].searchsorted(list(pivots), side="right")
        poses = (None,) + tuple(poses) + (None,)
        for i in range(op.n_partition):
            values = a.iloc[poses[i] : poses[i + 1]]
            ctx[out.key, (i,)] = values

    @classmethod
    def _execute_map(cls, ctx, op):
        a = [ctx[c.key] for c in op.inputs][0]
        if op.sort_type == "sort_values":
            if len(a.shape) == 2:
                # DataFrame type
                cls._execute_dataframe_map(ctx, op)
            else:
                # Series type
                cls._execute_series_map(ctx, op)
        else:
            cls._execute_sort_index_map(ctx, op)

    @classmethod
    def _execute_reduce(cls, ctx, op: "DataFramePSRSShuffle"):
        out_chunk = op.outputs[0]
        raw_inputs = list(op.iter_mapper_data(ctx, pop=False))

        xdf = pd if isinstance(raw_inputs[0], (pd.DataFrame, pd.Series)) else cudf
        if xdf is pd:
            concat_values = xdf.concat(raw_inputs, axis=op.axis, copy=False)
        else:
            concat_values = xdf.concat(raw_inputs, axis=op.axis)
        del raw_inputs[:]

        if isinstance(concat_values, xdf.DataFrame):
            concat_values.drop(
                _PSRS_DISTINCT_COL, axis=1, inplace=True, errors="ignore"
            )

            col_index_dtype = out_chunk.columns_value.to_pandas().dtype
            if concat_values.columns.dtype != col_index_dtype:
                concat_values.columns = concat_values.columns.astype(col_index_dtype)

        if op.sort_type == "sort_values":
            ctx[op.outputs[0].key] = execute_sort_values(concat_values, op)
        else:
            ctx[op.outputs[0].key] = execute_sort_index(concat_values, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        super().estimate_size(ctx, op)
        result = ctx[op.outputs[0].key]
        if op.stage == OperandStage.reduce:
            ctx[op.outputs[0].key] = (result[0], result[1] * 1.5)
        else:
            ctx[op.outputs[0].key] = result

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...config import Config, is_bool, is_list


task_options = Config()

# supervisor
task_options.register_option("optimize_tileable_graph", True, validator=is_bool)
task_options.register_option("optimize_chunk_graph", True, validator=is_bool)
task_options.register_option("fuse_enabled", True, validator=is_bool)

# worker
task_options.register_option("runtime_engines", ["numexpr", "cupy"], validator=is_list)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from enum import Enum
from string import ascii_letters, digits
from typing import Any, Optional, Dict

from ...core import TileableGraph
from ...serialization.serializables import (
    Serializable,
    StringField,
    ReferenceField,
    Int32Field,
    BoolField,
    AnyField,
    DictField,
    Float64Field,
)


class TaskStatus(Enum):
    pending = 0
    running = 1
    terminated = 2


class Task(Serializable):
    task_id: str = StringField("task_id")
    task_name: str = StringField("task_name")
    session_id: str = StringField("session_id")
    parent_task_id: str = StringField("parent_task_id")
    tileable_graph: TileableGraph = ReferenceField("tileable_graph", TileableGraph)
    fuse_enabled: bool = BoolField("fuse_enabled")
    rerun_time: int = Int32Field("rerun_time")
    extra_config: dict = DictField("extra_config")

    def __init__(
        self,
        task_id: str = None,
        session_id: str = None,
        tileable_graph: TileableGraph = None,
        task_name: str = None,
        parent_task_id: str = None,
        fuse_enabled: bool = True,
        rerun_time: int = 0,
        extra_config: dict = None,
    ):
        super().__init__(
            task_id=task_id,
            task_name=task_name,
            session_id=session_id,
            parent_task_id=parent_task_id,
            tileable_graph=tileable_graph,
            fuse_enabled=fuse_enabled,
            rerun_time=rerun_time,
            extra_config=extra_config,
        )


class TaskResult(Serializable):
    task_id: str = StringField("task_id")
    session_id: str = StringField("session_id")
    stage_id: str = StringField("stage_id")
    start_time: Optional[float] = Float64Field("start_time")
    end_time: Optional[float] = Float64Field("end_time")
    progress: Optional[float] = Float64Field("progress")
    status: TaskStatus = ReferenceField("status", TaskStatus)
    error = AnyField("error")
    traceback = AnyField("traceback")
    profiling: Dict = DictField("profiling")

    def __init__(
        self,
        task_id: str = None,
        session_id: str = None,
        stage_id: str = None,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        progress: Optional[float] = None,
        status: TaskStatus = None,
        error: Any = None,
        traceback: Any = None,
        profiling: Dict = None,
    ):
        super().__init__(
            task_id=task_id,
            session_id=session_id,
            stage_id=stage_id,
            start_time=start_time,
            end_time=end_time,
            progress=progress,
            status=status,
            error=error,
            traceback=traceback,
            profiling=profiling,
        )


def new_task_id():
    return "".join(random.choice(ascii_letters + digits) for _ in range(24))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple

import numpy as np

from .... import opcodes as OperandDef
from ....core import recursive_tile
from ....core.operand import OperandStage
from ....serialization.serializables import (
    FieldTypes,
    KeyField,
    AnyField,
    Float16Field,
    Int32Field,
    TupleField,
)
from ....utils import has_unknown_shape, require_module, ensure_own_data
from ....config import options
from ...operands import TensorMapReduceOperand, TensorOperandMixin, TensorShuffleProxy
from ...array_utils import as_same_device, device, cp
from ...core import TensorOrder
from ...datasource.array import tensor as astensor


class TensorPdist(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.PDIST

    _input = KeyField("input")
    _metric = AnyField("metric")
    _p = Float16Field("p", on_serialize=lambda x: float(x) if x is not None else x)
    _w = KeyField("w")
    _v = KeyField("V")
    _vi = KeyField("VI")
    _aggregate_size = Int32Field("aggregate_size")

    _a = KeyField("a")
    _a_offset = Int32Field("a_offset")
    _b = KeyField("b")
    _b_offset = Int32Field("b_offset")
    _out_sizes = TupleField("out_sizes", FieldTypes.int32)
    _n = Int32Field("n")

    def __init__(
        self,
        metric=None,
        p=None,
        w=None,
        v=None,
        vi=None,
        a=None,
        a_offset=None,
        b=None,
        b_offset=None,
        out_sizes=None,
        n=None,
        aggregate_size=None,
        **kw,
    ):
        super().__init__(
            _metric=metric,
            _p=p,
            _w=w,
            _v=v,
            _vi=vi,
            _a=a,
            _a_offset=a_offset,
            _b=b,
            _b_offset=b_offset,
            _out_sizes=out_sizes,
            _n=n,
            _aggregate_size=aggregate_size,
            **kw,
        )

    def _set_inputs(self, inputs: List) -> None:
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)

        if self.stage == OperandStage.map:
            self._a = next(inputs_iter)
            if self._b is not None:
                self._b = next(inputs_iter)
        else:
            self._input = next(inputs_iter)

        if self._w is not None:
            self._w = next(inputs_iter)
        if self._v is not None:
            self._v = next(inputs_iter)
        if self._vi is not None:
            self._vi = next(inputs_iter)

    @property
    def input(self):
        return self._input

    @property
    def metric(self):
        return self._metric

    @property
    def p(self):
        return self._p

    @property
    def w(self):
        return self._w

    @property
    def v(self):
        return self._v

    @property
    def vi(self):
        return self._vi

    @property
    def aggregate_size(self):
        return self._aggregate_size

    @property
    def a(self):
        return self._a

    @property
    def a_offset(self):
        return self._a_offset

    @property
    def b(self):
        return self._b

    @property
    def b_offset(self):
        return self._b_offset

    @property
    def out_sizes(self):
        return self._out_sizes

    @property
    def n(self):
        return self._n

    def __call__(self, x, shape: Tuple):
        inputs = [x]
        for val in [self._w, self._v, self._vi]:
            if val is not None:
                inputs.append(val)
        return self.new_tensor(inputs, shape=shape, order=TensorOrder.C_ORDER)

    @classmethod
    def _tile_one_chunk(cls, op, in_tensor, w, v, vi):
        out_tensor = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_inputs = [in_tensor.chunks[0]]
        for val in [w, v, vi]:
            if val is not None:
                chunk_inputs.append(val.chunks[0])
        chunk = chunk_op.new_chunk(
            chunk_inputs,
            shape=out_tensor.shape,
            order=out_tensor.order,
            index=(0,) * out_tensor.ndim,
        )

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out_tensor.shape,
            order=out_tensor.order,
            nsplits=tuple((s,) for s in out_tensor.shape),
            chunks=[chunk],
        )

    @classmethod
    def _tile_chunks(cls, op, in_tensor, w, v, vi):
        out_tensor = op.outputs[0]
        extra_inputs = []
        for val in [w, v, vi]:
            if val is not None:
                extra_inputs.append(val.chunks[0])

        n = in_tensor.shape[0]
        aggregate_size = op.aggregate_size
        if aggregate_size is None:
            aggregate_size = (
                np.ceil(
                    out_tensor.size
                    * out_tensor.dtype.itemsize
                    / options.chunk_store_limit
                )
                .astype(int)
                .item()
            )
        out_sizes = [out_tensor.size // aggregate_size for _ in range(aggregate_size)]
        for i in range(out_tensor.size % aggregate_size):
            out_sizes[i] += 1

        chunk_size = in_tensor.chunk_shape[0]
        map_chunks = []
        axis_0_cum_size = np.cumsum(in_tensor.nsplits[0]).tolist()
        for i in range(chunk_size):
            for j in range(i, chunk_size):
                kw = {
                    "stage": OperandStage.map,
                    "a": in_tensor.cix[i, 0],
                    "a_offset": axis_0_cum_size[i - 1] if i > 0 else 0,
                    "out_sizes": tuple(out_sizes),
                    "n": n,
                    "metric": op.metric,
                    "p": op.p,
                    "w": w.chunks[0] if w is not None else None,
                    "v": v.chunks[0] if v is not None else None,
                    "vi": vi.chunks[0] if vi is not None else None,
                    "dtype": out_tensor.dtype,
                }
                if i != j:
                    kw["b"] = in_tensor.cix[j, 0]
                    kw["b_offset"] = axis_0_cum_size[j - 1] if j > 0 else 0
                map_op = TensorPdist(**kw)
                map_chunk_inputs = [kw["a"]]
                if "b" in kw:
                    map_chunk_inputs.append(kw["b"])
                if kw["w"] is not None:
                    map_chunk_inputs.append(kw["w"])
                if kw["v"] is not None:
                    map_chunk_inputs.append(kw["v"])
                if kw["vi"] is not None:
                    map_chunk_inputs.append(kw["vi"])
                # calc chunk shape
                if i == j:
                    a_axis_0_size = kw["a"].shape[0]
                    chunk_shape = (a_axis_0_size * (a_axis_0_size - 1) // 2,)
                else:
                    chunk_shape = (kw["a"].shape[0] * kw["b"].shape[0],)
                map_chunk = map_op.new_chunk(
                    map_chunk_inputs,
                    shape=chunk_shape,
                    order=out_tensor.order,
                    index=(i * chunk_size + j,),
                )
                map_chunks.append(map_chunk)

        proxy_chunk = TensorShuffleProxy(dtype=out_tensor.dtype).new_chunk(
            map_chunks, shape=()
        )

        reduce_chunks = []
        for p in range(aggregate_size):
            reduce_chunk_op = TensorPdist(
                stage=OperandStage.reduce,
                dtype=out_tensor.dtype,
                reducer_ordinal=p,
                n_reducers=aggregate_size,
            )
            reduce_chunk = reduce_chunk_op.new_chunk(
                [proxy_chunk], shape=(out_sizes[p],), order=out_tensor.order, index=(p,)
            )
            reduce_chunks.append(reduce_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out_tensor.shape,
            order=out_tensor.order,
            nsplits=(tuple(out_sizes),),
            chunks=reduce_chunks,
        )

    @classmethod
    def tile(cls, op):
        # make sure every inputs have known shape
        if has_unknown_shape(*op.inputs):
            yield

        in_tensor = yield from recursive_tile(op.input.rechunk({1: op.input.shape[1]}))
        # rechunk w, v, vi into one chunk if any of them has value
        extra_inputs = [None] * 3
        for i, ei in enumerate([op.w, op.v, op.vi]):
            if ei is None:
                continue
            new_ei = yield from recursive_tile(ei.rechunk(ei.shape))
            extra_inputs[i] = new_ei
        w, v, vi = extra_inputs

        if len(in_tensor.chunks) == 1:
            # only 1 chunk
            return cls._tile_one_chunk(op, in_tensor, w, v, vi)
        else:
            return cls._tile_chunks(op, in_tensor, w, v, vi)

    @classmethod
    def _execute_map(cls, ctx, op):
        from scipy.spatial.distance import pdist, cdist

        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )

        if xp is cp:  # pragma: no cover
            raise NotImplementedError("`pdist` does not support running on GPU yet")

        with device(device_id):
            inputs_iter = iter(inputs)
            a = next(inputs_iter)
            if op.b is not None:
                b = next(inputs_iter)
            else:
                b = None
            kw = dict()
            if op.p is not None:
                kw["p"] = op.p
            if op.w is not None:
                kw["w"] = next(inputs_iter)
            if op.v is not None:
                kw["V"] = next(inputs_iter)
            if op.vi is not None:
                kw["VI"] = next(inputs_iter)
            metric = op.metric

            if b is None:
                # one input, pdist on same chunk
                dists = pdist(ensure_own_data(a), metric=metric, **kw)
                i_indices, j_indices = xp.triu_indices(a.shape[0], k=1)
                i_indices += op.a_offset
                j_indices += op.a_offset
            else:
                # two inputs, pdist on different chunks
                dists = cdist(
                    ensure_own_data(a), ensure_own_data(b), metric=metric, **kw
                ).ravel()
                mgrid = xp.mgrid[
                    op.a_offset : op.a_offset + a.shape[0],
                    op.b_offset : op.b_offset + b.shape[0],
                ]
                i_indices, j_indices = mgrid[0].ravel(), mgrid[1].ravel()

            out_row_sizes = xp.arange(op.n - 1, -1, -1)
            out_row_cum_sizes = xp.empty((op.n + 1,), dtype=int)
            out_row_cum_sizes[0] = 0
            xp.cumsum(out_row_sizes, out=out_row_cum_sizes[1:])
            indices = (
                out_row_cum_sizes[i_indices]
                + j_indices
                - (op.n - out_row_sizes[i_indices])
            )

            # save as much memory as possible
            del i_indices, j_indices, out_row_sizes, out_row_cum_sizes

            out_cum_size = xp.cumsum(op.out_sizes)
            out = op.outputs[0]
            for i in range(len(op.out_sizes)):
                start_index = out_cum_size[i - 1] if i > 0 else 0
                end_index = out_cum_size[i]
                to_filter = (indices >= start_index) & (indices < end_index)
                downside_indices = indices[to_filter] - start_index
                downside_dists = dists[to_filter]
                ctx[out.key, (i,)] = (downside_indices, downside_dists)

    @classmethod
    def _execute_single(cls, ctx, op):
        from scipy.spatial.distance import pdist

        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )

        if xp is cp:  # pragma: no cover
            raise NotImplementedError("`pdist` does not support running on GPU yet")

        with device(device_id):
            inputs_iter = iter(inputs)
            x = next(inputs_iter)
            kw = dict()
            if op.p is not None:
                kw["p"] = op.p
            if op.w is not None:
                kw["w"] = next(inputs_iter)
            if op.v is not None:
                kw["V"] = next(inputs_iter)
            if op.vi is not None:
                kw["VI"] = next(inputs_iter)

        ctx[op.outputs[0].key] = pdist(ensure_own_data(x), metric=op.metric, **kw)

    @classmethod
    def _execute_reduce(cls, ctx, op: "TensorPdist"):
        raw_inputs = list(op.iter_mapper_data(ctx))
        raw_indices = [inp[0] for inp in raw_inputs]
        raw_dists = [inp[1] for inp in raw_inputs]
        inputs, device_id, xp = as_same_device(
            raw_indices + raw_dists, op.device, ret_extra=True
        )
        raw_indices = inputs[: len(raw_indices)]
        raw_dists = inputs[len(raw_indices) :]
        output = op.outputs[0]

        with device(device_id):
            indices = xp.concatenate(raw_indices)
            dists = xp.concatenate(raw_dists)
            out_dists = xp.empty(output.shape, dtype=float)
            out_dists[indices] = dists
            ctx[output.key] = out_dists

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls._execute_reduce(ctx, op)
        else:
            cls._execute_single(ctx, op)


@require_module("scipy.spatial.distance")
def pdist(X, metric="euclidean", **kwargs):
    """
    Pairwise distances between observations in n-dimensional space.

    See Notes for common calling conventions.

    Parameters
    ----------
    X : Tensor
        An m by n tensor of m original observations in an
        n-dimensional space.
    metric : str or function, optional
        The distance metric to use. The distance function can
        be 'braycurtis', 'canberra', 'chebyshev', 'cityblock',
        'correlation', 'cosine', 'dice', 'euclidean', 'hamming',
        'jaccard', 'jensenshannon', 'kulsinski', 'mahalanobis', 'matching',
        'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean',
        'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule'.
    **kwargs : dict, optional
        Extra arguments to `metric`: refer to each metric documentation for a
        list of all possible arguments.

        Some possible arguments:

        p : scalar
        The p-norm to apply for Minkowski, weighted and unweighted.
        Default: 2.

        w : Tensor
        The weight vector for metrics that support weights (e.g., Minkowski).

        V : Tensor
        The variance vector for standardized Euclidean.
        Default: var(X, axis=0, ddof=1)

        VI : Tensor
        The inverse of the covariance matrix for Mahalanobis.
        Default: inv(cov(X.T)).T

        out : Tensor.
        The output tensor
        If not None, condensed distance matrix Y is stored in this tensor.
        Note: metric independent, it will become a regular keyword arg in a
        future scipy version

    Returns
    -------
    Y : Tensor
        Returns a condensed distance matrix Y.  For
        each :math:`i` and :math:`j` (where :math:`i<j<m`),where m is the number
        of original observations. The metric ``dist(u=X[i], v=X[j])``
        is computed and stored in entry ``ij``.

    See Also
    --------
    squareform : converts between condensed distance matrices and
                 square distance matrices.

    Notes
    -----
    See ``squareform`` for information on how to calculate the index of
    this entry or to convert the condensed distance matrix to a
    redundant square matrix.

    The following are common calling conventions.

    1. ``Y = pdist(X, 'euclidean')``

       Computes the distance between m points using Euclidean distance
       (2-norm) as the distance metric between the points. The points
       are arranged as m n-dimensional row vectors in the matrix X.

    2. ``Y = pdist(X, 'minkowski', p=2.)``

       Computes the distances using the Minkowski distance
       :math:`||u-v||_p` (p-norm) where :math:`p \\geq 1`.

    3. ``Y = pdist(X, 'cityblock')``

       Computes the city block or Manhattan distance between the
       points.

    4. ``Y = pdist(X, 'seuclidean', V=None)``

       Computes the standardized Euclidean distance. The standardized
       Euclidean distance between two n-vectors ``u`` and ``v`` is

       .. math::

          \\sqrt{\\sum {(u_i-v_i)^2 / V[x_i]}}


       V is the variance vector; V[i] is the variance computed over all
       the i'th components of the points.  If not passed, it is
       automatically computed.

    5. ``Y = pdist(X, 'sqeuclidean')``

       Computes the squared Euclidean distance :math:`||u-v||_2^2` between
       the vectors.

    6. ``Y = pdist(X, 'cosine')``

       Computes the cosine distance between vectors u and v,

       .. math::

          1 - \\frac{u \\cdot v}
                   {{||u||}_2 {||v||}_2}

       where :math:`||*||_2` is the 2-norm of its argument ``*``, and
       :math:`u \\cdot v` is the dot product of ``u`` and ``v``.

    7. ``Y = pdist(X, 'correlation')``

       Computes the correlation distance between vectors u and v. This is

       .. math::

          1 - \\frac{(u - \\bar{u}) \\cdot (v - \\bar{v})}
                   {{||(u - \\bar{u})||}_2 {||(v - \\bar{v})||}_2}

       where :math:`\\bar{v}` is the mean of the elements of vector v,
       and :math:`x \\cdot y` is the dot product of :math:`x` and :math:`y`.

    8. ``Y = pdist(X, 'hamming')``

       Computes the normalized Hamming distance, or the proportion of
       those vector elements between two n-vectors ``u`` and ``v``
       which disagree. To save memory, the matrix ``X`` can be of type
       boolean.

    9. ``Y = pdist(X, 'jaccard')``

       Computes the Jaccard distance between the points. Given two
       vectors, ``u`` and ``v``, the Jaccard distance is the
       proportion of those elements ``u[i]`` and ``v[i]`` that
       disagree.

    10. ``Y = pdist(X, 'chebyshev')``

       Computes the Chebyshev distance between the points. The
       Chebyshev distance between two n-vectors ``u`` and ``v`` is the
       maximum norm-1 distance between their respective elements. More
       precisely, the distance is given by

       .. math::

          d(u,v) = \\max_i {|u_i-v_i|}

    11. ``Y = pdist(X, 'canberra')``

       Computes the Canberra distance between the points. The
       Canberra distance between two points ``u`` and ``v`` is

       .. math::

         d(u,v) = \\sum_i \\frac{|u_i-v_i|}
                              {|u_i|+|v_i|}


    12. ``Y = pdist(X, 'braycurtis')``

       Computes the Bray-Curtis distance between the points. The
       Bray-Curtis distance between two points ``u`` and ``v`` is


       .. math::

            d(u,v) = \\frac{\\sum_i {|u_i-v_i|}}
                           {\\sum_i {|u_i+v_i|}}

    13. ``Y = pdist(X, 'mahalanobis', VI=None)``

       Computes the Mahalanobis distance between the points. The
       Mahalanobis distance between two points ``u`` and ``v`` is
       :math:`\\sqrt{(u-v)(1/V)(u-v)^T}` where :math:`(1/V)` (the ``VI``
       variable) is the inverse covariance. If ``VI`` is not None,
       ``VI`` will be used as the inverse covariance matrix.

    14. ``Y = pdist(X, 'yule')``

       Computes the Yule distance between each pair of boolean
       vectors. (see yule function documentation)

    15. ``Y = pdist(X, 'matching')``

       Synonym for 'hamming'.

    16. ``Y = pdist(X, 'dice')``

       Computes the Dice distance between each pair of boolean
       vectors. (see dice function documentation)

    17. ``Y = pdist(X, 'kulsinski')``

       Computes the Kulsinski distance between each pair of
       boolean vectors. (see kulsinski function documentation)

    18. ``Y = pdist(X, 'rogerstanimoto')``

       Computes the Rogers-Tanimoto distance between each pair of
       boolean vectors. (see rogerstanimoto function documentation)

    19. ``Y = pdist(X, 'russellrao')``

       Computes the Russell-Rao distance between each pair of
       boolean vectors. (see russellrao function documentation)

    20. ``Y = pdist(X, 'sokalmichener')``

       Computes the Sokal-Michener distance between each pair of
       boolean vectors. (see sokalmichener function documentation)

    21. ``Y = pdist(X, 'sokalsneath')``

       Computes the Sokal-Sneath distance between each pair of
       boolean vectors. (see sokalsneath function documentation)

    22. ``Y = pdist(X, 'wminkowski', p=2, w=w)``

       Computes the weighted Minkowski distance between each pair of
       vectors. (see wminkowski function documentation)

    23. ``Y = pdist(X, f)``

       Computes the distance between all pairs of vectors in X
       using the user supplied 2-arity function f. For example,
       Euclidean distance between the vectors could be computed
       as follows::

         dm = pdist(X, lambda u, v: np.sqrt(((u-v)**2).sum()))

       Note that you should avoid passing a reference to one of
       the distance functions defined in this library. For example,::

         dm = pdist(X, sokalsneath)

       would calculate the pair-wise distances between the vectors in
       X using the Python function sokalsneath. This would result in
       sokalsneath being called :math:`{n \\choose 2}` times, which
       is inefficient. Instead, the optimized C version is more
       efficient, and we call it using the following syntax.::

         dm = pdist(X, 'sokalsneath')

    """

    X = astensor(X, order="C")

    if X.issparse():
        raise ValueError("Sparse tensors are not supported by this function.")

    s = X.shape
    if len(s) != 2:
        raise ValueError("A 2-dimensional tensor must be passed.")

    m = s[0]
    out = kwargs.pop("out", None)
    if out is not None:
        if not hasattr(out, "shape"):
            raise TypeError("return arrays must be a tensor")
        if out.shape != (m * (m - 1) // 2,):
            raise ValueError("output tensor has incorrect shape.")
        if out.dtype != np.double:
            raise ValueError("Output tensor must be double type.")

    if not callable(metric) and not isinstance(metric, str):
        raise TypeError(
            "2nd argument metric must be a string identifier or a function."
        )

    # scipy remove "wminkowski" since v1.8.0, use "minkowski" with `w=`
    # keyword-argument for the given weight.
    if metric == "wminkowski":
        metric = "minkowski"

    p = kwargs.pop("p", None)
    w = kwargs.pop("w", None)
    if w is not None:
        w = astensor(w)
    v = kwargs.pop("V", None)
    if v is not None:
        v = astensor(v)
    vi = kwargs.pop("VI", None)
    if vi is not None:
        vi = astensor(vi)
    aggregate_size = kwargs.pop("aggregate_size", None)

    if len(kwargs) > 0:
        raise TypeError(
            f"`pdist` got an unexpected keyword argument '{next(iter(kwargs))}'"
        )

    op = TensorPdist(
        metric=metric,
        p=p,
        w=w,
        v=v,
        vi=vi,
        aggregate_size=aggregate_size,
        dtype=np.dtype(float),
    )
    shape = (m * (m - 1) // 2,)
    ret = op(X, shape)

    if out is None:
        return ret
    else:
        out.data = ret.data
        return out



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType, recursive_tile
from ...serialization.serializables import KeyField, StringField, TupleField, DictField
from ...tensor import tensor as astensor
from ...tensor.core import TENSOR_TYPE
from ...utils import has_unknown_shape
from ..align import align_series_series
from ..core import SERIES_TYPE
from ..initializer import Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_series, parse_index, infer_index_value


class SeriesStringMethod(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.STRING_METHOD

    _input = KeyField("input")
    _method = StringField("method")
    _method_args = TupleField("method_args")
    _method_kwargs = DictField("method_kwargs")

    def __init__(
        self, method=None, method_args=None, method_kwargs=None, output_types=None, **kw
    ):
        super().__init__(
            _method=method,
            _method_args=method_args,
            _method_kwargs=method_kwargs,
            _output_types=output_types,
            **kw
        )
        if not self.output_types:
            self.output_types = [OutputType.series]

    @property
    def input(self):
        return self._input

    @property
    def method(self):
        return self._method

    @property
    def method_args(self):
        return self._method_args

    @property
    def method_kwargs(self):
        return self._method_kwargs

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(self._inputs) == 2:
            # for method cat
            self._method_kwargs["others"] = self._inputs[1]

    def __call__(self, inp):
        return _string_method_to_handlers[self._method].call(self, inp)

    @classmethod
    def tile(cls, op):
        tiled = _string_method_to_handlers[op.method].tile(op)
        if inspect.isgenerator(tiled):
            return (yield from tiled)
        else:
            return tiled

    @classmethod
    def execute(cls, ctx, op):
        return _string_method_to_handlers[op.method].execute(ctx, op)


class SeriesStringMethodBaseHandler:
    @classmethod
    def call(cls, op, inp):
        empty_series = build_empty_series(inp.dtype)
        dtype = getattr(empty_series.str, op.method)(
            *op.method_args, **op.method_kwargs
        ).dtype
        return op.new_series(
            [inp],
            shape=inp.shape,
            dtype=dtype,
            index_value=inp.index_value,
            name=inp.name,
        )

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]
        out_chunks = []
        for series_chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [series_chunk],
                shape=series_chunk.shape,
                dtype=out.dtype,
                index=series_chunk.index,
                index_value=series_chunk.index_value,
                name=series_chunk.name,
            )
            out_chunks.append(out_chunk)

        params = out.params
        params["chunks"] = out_chunks
        params["nsplits"] = op.input.nsplits
        new_op = op.copy()
        return new_op.new_tileables([op.input], kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = getattr(inp.str, op.method)(
            *op.method_args, **op.method_kwargs
        )


class SeriesStringSplitHandler(SeriesStringMethodBaseHandler):
    @classmethod
    def call(cls, op, inp):
        method_kwargs = op.method_kwargs
        if method_kwargs.get("expand", False) is False:
            return super().call(op, inp)
        n = method_kwargs.get("n", -1)
        # does not support if expand and n == -1
        if n == -1:  # pragma: no cover
            raise NotImplementedError("`n` needs to be specified when expand=True")

        op.output_types = [OutputType.dataframe]
        columns = pd.RangeIndex(n + 1)
        columns_value = parse_index(columns, store_data=True)
        dtypes = pd.Series([inp.dtype] * len(columns), index=columns)
        return op.new_dataframe(
            [inp],
            shape=(inp.shape[0], len(columns)),
            dtypes=dtypes,
            columns_value=columns_value,
            index_value=inp.index_value,
        )

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]

        if out.op.output_types[0] == OutputType.series:
            return super().tile(op)

        out_chunks = []
        columns = out.columns_value.to_pandas()
        for series_chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [series_chunk],
                shape=(series_chunk.shape[0], len(columns)),
                index=(series_chunk.index[0], 0),
                dtypes=out.dtypes,
                index_value=series_chunk.index_value,
                columns_value=out.columns_value,
            )
            out_chunks.append(out_chunk)

        params = out.params
        params["chunks"] = out_chunks
        params["nsplits"] = (op.input.nsplits[0], (len(columns),))
        new_op = op.copy()
        return new_op.new_tileables([op.input], kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.input.key]
        out = op.outputs[0]
        result = getattr(inp.str, op.method)(*op.method_args, **op.method_kwargs)
        if result.ndim == 2 and result.shape[1] < out.shape[1]:
            for i in range(result.shape[1], out.shape[1]):
                result[i] = None
        ctx[op.outputs[0].key] = result


class SeriesStringCatHandler(SeriesStringMethodBaseHandler):
    CAT_TYPE_ERROR = (
        "others must be Series, Index, DataFrame, "
        "Tensor, np.ndarrary or list-like "
        "(either containing only strings or "
        "containing only objects of "
        "type Series/Index/Tensor/np.ndarray[1-dim])"
    )
    CAT_LEN_ERROR = (
        "If `others` contains arrays or lists (or other list-likes without an index), "
        "these must all be of the same length as the calling Series/Index."
    )

    @classmethod
    def call(cls, op, inp):
        method_kwargs = op.method_kwargs
        others = method_kwargs.get("others")

        if others is None:
            from ..reduction import build_str_concat_object

            return build_str_concat_object(
                inp,
                sep=op.method_kwargs.get("sep"),
                na_rep=op.method_kwargs.get("na_rep"),
            )
        elif isinstance(others, (tuple, list, np.ndarray, TENSOR_TYPE)):
            others = astensor(others, dtype=object)
            if others.ndim != 1:
                raise TypeError(cls.CAT_TYPE_ERROR)
            if (
                not np.isnan(inp.shape[0])
                and not np.isnan(others.shape[0])
                and inp.shape[0] != others.shape[0]
            ):
                raise ValueError(cls.CAT_LEN_ERROR)
            inputs = [inp]
            if isinstance(others, TENSOR_TYPE):
                inputs.append(others)
            return op.new_series(
                inputs,
                shape=inp.shape,
                dtype=inp.dtype,
                index_value=inp.index_value,
                name=inp.name,
            )
        elif isinstance(others, (pd.Series, SERIES_TYPE)):
            others = asseries(others)
            if op.method_kwargs.get("join") != "outer":  # pragma: no cover
                raise NotImplementedError("only outer join supported for now")
            return op.new_series(
                [inp, others],
                shape=inp.shape,
                dtype=inp.dtype,
                index_value=infer_index_value(inp.index_value, others.index_value),
                name=inp.name,
            )
        elif isinstance(others, str) and op.method_kwargs.get("sep") is None:
            raise ValueError("Did you mean to supply a `sep` keyword?")
        else:
            raise TypeError(cls.CAT_TYPE_ERROR)

    @classmethod
    def tile(cls, op):
        inp = op.input
        out = op.outputs[0]

        # aggregation concat resulting in scalars is redirected
        assert out.ndim != 0

        if isinstance(op.inputs[1], TENSOR_TYPE):
            if has_unknown_shape(*op.inputs):
                yield
            # rechunk others as input
            others = yield from recursive_tile(op.inputs[1].rechunk(op.input.nsplits))
            out_chunks = []
            for c in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op._method_kwargs = op.method_kwargs.copy()
                out_chunk = chunk_op.new_chunk(
                    [c, others.cix[c.index]],
                    dtype=c.dtype,
                    index=c.index,
                    shape=c.shape,
                    index_value=c.index_value,
                    name=c.name,
                )
                out_chunks.append(out_chunk)
            new_op = op.copy()
            params = out.params
            params["nsplits"] = inp.nsplits
            params["chunks"] = out_chunks
            return new_op.new_tileables(op.inputs, kws=[params])
        elif isinstance(op.inputs[1], SERIES_TYPE):
            # both series
            out_chunks = []
            nsplits, _, left_chunks, right_chunks = align_series_series(*op.inputs)
            for left_chunk, right_chunk in zip(left_chunks, right_chunks):
                chunk_op = op.copy().reset_key()
                chunk_op._method_kwargs = op.method_kwargs.copy()
                params = left_chunk.params
                params["name"] = out.name
                out_chunk = chunk_op.new_chunk([left_chunk, right_chunk], **params)
                out_chunks.append(out_chunk)
            new_op = op.copy()
            params = out.params
            params["nsplits"] = nsplits
            params["chunks"] = out_chunks
            return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inputs = [ctx[inp.key] for inp in op.inputs]
        method_kwargs = op.method_kwargs

        # aggregation concat is redirected and `others` is always defined
        assert len(inputs) > 1

        method_kwargs["others"] = inputs[1]
        ctx[op.outputs[0].key] = inputs[0].str.cat(**method_kwargs)


class SeriesStringExtractHandler(SeriesStringMethodBaseHandler):
    @classmethod
    def call(cls, op, inp):
        empty_series = build_empty_series(
            inp.dtype, index=inp.index_value.to_pandas()[:0]
        )
        test_df = getattr(empty_series.str, op.method)(
            *op.method_args, **op.method_kwargs
        )
        if test_df.ndim == 1:
            return op.new_series(
                [inp],
                shape=inp.shape,
                dtype=test_df.dtype,
                index_value=inp.index_value,
                name=inp.name,
            )
        else:
            op.output_types = [OutputType.dataframe]
            if op.method == "extractall":
                index_value = parse_index(test_df.index, inp)
                shape = (np.nan, test_df.shape[1])
            else:
                index_value = inp.index_value
                shape = (inp.shape[0], test_df.shape[1])
            return op.new_dataframe(
                [inp],
                shape=shape,
                dtypes=test_df.dtypes,
                index_value=index_value,
                columns_value=parse_index(test_df.columns, store_data=True),
            )

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]
        out_chunks = []
        for series_chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            if out.ndim == 1:
                out_chunk = chunk_op.new_chunk(
                    [series_chunk],
                    shape=series_chunk.shape,
                    index=series_chunk.index,
                    dtype=out.dtype,
                    index_value=series_chunk.index_value,
                    name=out.name,
                )
            else:
                if op.method == "extract":
                    index_value = series_chunk.index_value
                    shape = (series_chunk.shape[0], out.shape[1])
                else:
                    index_value = parse_index(
                        out.index_value.to_pandas()[:0], series_chunk
                    )
                    shape = (np.nan, out.shape[1])
                out_chunk = chunk_op.new_chunk(
                    [series_chunk],
                    shape=shape,
                    index=(series_chunk.index[0], 0),
                    dtypes=out.dtypes,
                    index_value=index_value,
                    columns_value=out.columns_value,
                )
            out_chunks.append(out_chunk)

        out = op.outputs[0]
        params = out.params
        params["chunks"] = out_chunks
        if out.ndim == 1:
            params["nsplits"] = op.input.nsplits
        elif op.method == "extract":
            params["nsplits"] = (op.input.nsplits[0], (out.shape[1],))
        else:
            params["nsplits"] = ((np.nan,) * len(op.input.nsplits[0]), (out.shape[1],))
        new_op = op.copy()
        return new_op.new_tileables([op.input], kws=[params])


_string_method_to_handlers = {}
_not_implements = ["get_dummies"]
# start to register handlers for string methods
# register special methods first
_string_method_to_handlers["split"] = SeriesStringSplitHandler
_string_method_to_handlers["rsplit"] = SeriesStringSplitHandler
_string_method_to_handlers["cat"] = SeriesStringCatHandler
_string_method_to_handlers["extract"] = SeriesStringExtractHandler
_string_method_to_handlers["extractall"] = SeriesStringExtractHandler
# then come to the normal methods
for method in dir(pd.Series.str):
    if method.startswith("_") and method != "__getitem__":
        continue
    if method in _not_implements:
        continue
    if method in _string_method_to_handlers:
        continue
    _string_method_to_handlers[method] = SeriesStringMethodBaseHandler



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import AsyncGenerator, Dict, List, Optional, Type

from ..core import NodeRole


class AbstractClusterBackend(ABC):
    name = None

    @classmethod
    @abstractmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: Optional[str], pool_address: str
    ) -> "AbstractClusterBackend":
        """

        Parameters
        ----------
        node_role
        lookup_address
        pool_address

        Returns
        -------

        """

    @abstractmethod
    async def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        """
        Watch changes of supervisors

        Returns
        -------
        out : AsyncGenerator[List[str]]
            Generator of list of schedulers
        """

    @abstractmethod
    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        """
        Get list of supervisors

        Parameters
        ----------
        filter_ready : bool
            True if return ready nodes only, or return starting and ready nodes

        Returns
        -------
        out : List[str]
            List of supervisors
        """

    @abstractmethod
    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        """
        Create a new worker

        Returns
        -------
        Address of the new created worker
        """

    @abstractmethod
    async def release_worker(self, address: str):
        """
        Return a worker
        """

    @abstractmethod
    async def reconstruct_worker(self, address: str):
        """
        Reconstruct a worker
        """


_cluster_backend_types: Dict[str, Type[AbstractClusterBackend]] = dict()


def register_cluster_backend(backend: Type[AbstractClusterBackend]):
    _cluster_backend_types[backend.name] = backend
    return backend


def get_cluster_backend(backend_name: str) -> Type[AbstractClusterBackend]:
    return _cluster_backend_types[backend_name]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import atexit
import os
import sys
import logging
from concurrent.futures import Future as SyncFuture
from typing import Dict, List, Union

import numpy as np

from ... import oscar as mo
from ...core.entrypoints import init_extension_entrypoints
from ...lib.aio import get_isolation, stop_isolation
from ...oscar.backends.router import Router
from ...resource import cpu_count, cuda_count, mem_total
from ...services import NodeRole
from ...services.task.execution.api import ExecutionConfig
from ...typing import ClusterType, ClientType
from ..utils import get_third_party_modules_from_config, load_config
from .pool import create_supervisor_actor_pool, create_worker_actor_pool
from .service import (
    start_supervisor,
    start_worker,
    stop_supervisor,
    stop_worker,
)
from .session import AbstractSession, _new_session, ensure_isolation_created

logger = logging.getLogger(__name__)

_is_exiting_future = SyncFuture()
atexit.register(
    lambda: _is_exiting_future.set_result(0) if not _is_exiting_future.done() else None
)
atexit.register(stop_isolation)

# The default config file.
DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "config.yml"
)

# the default times to retry subtask.
DEFAULT_SUBTASK_MAX_RETRIES = 3
# the default time to cancel a subtask.
DEFAULT_SUBTASK_CANCEL_TIMEOUT = 5


def _load_config(config: Union[str, Dict] = None):
    return load_config(config, default_config_file=DEFAULT_CONFIG_FILE)


async def new_cluster_in_isolation(
    address: str = "0.0.0.0",
    n_worker: int = 1,
    n_cpu: Union[int, str] = "auto",
    mem_bytes: Union[int, str] = "auto",
    cuda_devices: Union[List[int], str] = "auto",
    subprocess_start_method: str = None,
    backend: str = None,
    config: Union[str, Dict] = None,
    web: bool = True,
    timeout: float = None,
    n_supervisor_process: int = 0,
) -> ClientType:
    cluster = LocalCluster(
        address,
        n_worker,
        n_cpu,
        mem_bytes,
        cuda_devices,
        subprocess_start_method,
        backend,
        config,
        web,
        n_supervisor_process,
    )
    await cluster.start()
    return await LocalClient.create(cluster, timeout)


async def new_cluster(
    address: str = "0.0.0.0",
    n_worker: int = 1,
    n_cpu: Union[int, str] = "auto",
    mem_bytes: Union[int, str] = "auto",
    cuda_devices: Union[List[int], str] = "auto",
    subprocess_start_method: str = None,
    backend: str = None,
    config: Union[str, Dict] = None,
    web: bool = True,
    loop: asyncio.AbstractEventLoop = None,
    use_uvloop: Union[bool, str] = "auto",
    n_supervisor_process: int = 0,
) -> ClientType:
    coro = new_cluster_in_isolation(
        address,
        n_worker=n_worker,
        n_cpu=n_cpu,
        mem_bytes=mem_bytes,
        cuda_devices=cuda_devices,
        subprocess_start_method=subprocess_start_method,
        backend=backend,
        config=config,
        web=web,
        n_supervisor_process=n_supervisor_process,
    )
    isolation = ensure_isolation_created(dict(loop=loop, use_uvloop=use_uvloop))
    fut = asyncio.run_coroutine_threadsafe(coro, isolation.loop)
    client = await asyncio.wrap_future(fut)
    client.session.as_default()
    return client


async def stop_cluster(cluster: ClusterType):
    isolation = get_isolation()
    coro = cluster.stop()
    await asyncio.wrap_future(asyncio.run_coroutine_threadsafe(coro, isolation.loop))
    Router.set_instance(None)


class LocalCluster:
    def __init__(
        self: ClusterType,
        address: str = "0.0.0.0",
        n_worker: int = 1,
        n_cpu: Union[int, str] = "auto",
        mem_bytes: Union[int, str] = "auto",
        cuda_devices: Union[List[int], List[List[int]], str] = "auto",
        subprocess_start_method: str = None,
        backend: str = None,
        config: Union[str, Dict] = None,
        web: Union[bool, str] = "auto",
        n_supervisor_process: int = 0,
    ):
        # load third party extensions.
        init_extension_entrypoints()
        # auto choose the subprocess_start_method.
        if subprocess_start_method is None:
            subprocess_start_method = (
                "spawn" if sys.platform == "win32" else "forkserver"
            )
        self._address = address
        self._n_worker = n_worker
        self._n_cpu = cpu_count() if n_cpu == "auto" else n_cpu
        self._mem_bytes = mem_total() if mem_bytes == "auto" else mem_bytes
        self._cuda_devices = self._get_cuda_devices(cuda_devices, n_worker)
        self._subprocess_start_method = subprocess_start_method
        self._config = load_config(config, default_config_file=DEFAULT_CONFIG_FILE)
        execution_config = ExecutionConfig.from_config(self._config, backend=backend)
        self._backend = execution_config.backend
        self._web = web
        self._n_supervisor_process = n_supervisor_process

        execution_config.merge_from(
            ExecutionConfig.from_params(
                backend=self._backend,
                n_worker=self._n_worker,
                n_cpu=self._n_cpu,
                mem_bytes=self._mem_bytes,
                cuda_devices=self._cuda_devices,
                subtask_cancel_timeout=self._config.get("scheduling", {}).get(
                    "subtask_cancel_timeout", DEFAULT_SUBTASK_CANCEL_TIMEOUT
                ),
                subtask_max_retries=self._config.get("scheduling", {}).get(
                    "subtask_max_retries", DEFAULT_SUBTASK_MAX_RETRIES
                ),
            )
        )

        self._bands_to_resource = execution_config.get_deploy_band_resources()
        self._supervisor_pool = None
        self._worker_pools = []
        self._exiting_check_task = None

        self.supervisor_address = None
        self.web_address = None

    @staticmethod
    def _get_cuda_devices(cuda_devices, n_worker):
        if cuda_devices == "auto":
            total = cuda_count()
            all_devices = np.arange(total)
            return [list(arr) for arr in np.array_split(all_devices, n_worker)]

        else:  # pragma: no cover
            if isinstance(cuda_devices[0], int):
                assert n_worker == 1
                return [cuda_devices]
            else:
                assert len(cuda_devices) == n_worker
                return cuda_devices

    @property
    def backend(self):
        return self._backend

    @property
    def external_address(self):
        return self._supervisor_pool.external_address

    async def start(self):
        await self._start_supervisor_pool()
        await self._start_worker_pools()
        # start service
        await self._start_service()

        if self._web:
            from ...services.web.supervisor import WebActor

            web_actor = await mo.actor_ref(
                WebActor.default_uid(), address=self.supervisor_address
            )
            self.web_address = await web_actor.get_web_address()
            logger.warning("Web service started at %s", self.web_address)

        self._exiting_check_task = asyncio.create_task(self._check_exiting())

    async def _check_exiting(self):
        await asyncio.wrap_future(_is_exiting_future)
        await self.stop()

    async def _start_supervisor_pool(self):
        supervisor_modules = get_third_party_modules_from_config(
            self._config, NodeRole.SUPERVISOR
        )
        self._supervisor_pool = await create_supervisor_actor_pool(
            self._address,
            n_process=self._n_supervisor_process,
            modules=supervisor_modules,
            subprocess_start_method=self._subprocess_start_method,
            metrics=self._config.get("metrics", {}),
        )
        self.supervisor_address = self._supervisor_pool.external_address

    async def _start_worker_pools(self):
        worker_modules = get_third_party_modules_from_config(
            self._config, NodeRole.WORKER
        )
        for band_to_resource in self._bands_to_resource:
            worker_pool = await create_worker_actor_pool(
                self._address,
                band_to_resource,
                modules=worker_modules,
                subprocess_start_method=self._subprocess_start_method,
                metrics=self._config.get("metrics", {}),
            )
            self._worker_pools.append(worker_pool)

    async def _start_service(self):
        self._web = await start_supervisor(
            self.supervisor_address, config=self._config, web=self._web
        )
        for worker_pool, band_to_resource in zip(
            self._worker_pools, self._bands_to_resource
        ):
            await start_worker(
                worker_pool.external_address,
                self.supervisor_address,
                band_to_resource,
                config=self._config,
            )

    async def stop(self):
        from .session import SessionAPI

        # delete all sessions
        session_api = await SessionAPI.create(self._supervisor_pool.external_address)
        await session_api.delete_all_sessions()

        for worker_pool in self._worker_pools:
            await stop_worker(worker_pool.external_address, self._config)
        await stop_supervisor(self._supervisor_pool.external_address, self._config)
        for worker_pool in self._worker_pools:
            await worker_pool.stop()
        await self._supervisor_pool.stop()
        AbstractSession.reset_default()
        self._exiting_check_task.cancel()
        Router.set_instance(None)


class LocalClient:
    def __init__(self: ClientType, cluster: ClusterType, session: AbstractSession):
        self._cluster = cluster
        self.session = session

    @classmethod
    async def create(
        cls,
        cluster: LocalCluster,
        timeout: float = None,
    ) -> ClientType:
        session = await _new_session(
            cluster.external_address,
            backend=cluster.backend,
            default=True,
            timeout=timeout,
        )
        client = LocalClient(cluster, session)
        session.client = client
        return client

    @property
    def web_address(self):
        return self._cluster.web_address

    async def __aenter__(self):
        return self

    async def __aexit__(self, *_):
        await self.stop()

    async def stop(self):
        await stop_cluster(self._cluster)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import KeyField, BoolField, Int64Field, StringField
from ...utils import has_unknown_shape, pd_release_version
from ..core import Series
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_series, parse_index

_keep_original_order = pd_release_version >= (1, 3, 0)


class DataFrameValueCounts(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.VALUE_COUNTS

    input = KeyField("input")
    normalize = BoolField("normalize")
    sort = BoolField("sort")
    ascending = BoolField("ascending")
    bins = Int64Field("bins")
    dropna = BoolField("dropna")
    method = StringField("method")
    convert_index_to_interval = BoolField("convert_index_to_interval", default=None)
    nrows = Int64Field("nrows", default=None)

    def __init__(self, **kw):
        super().__init__(**kw)
        self.output_types = [OutputType.series]

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self.input = self._inputs[0]

    def __call__(self, inp):
        test_series = build_series(inp).value_counts(normalize=self.normalize)
        if self.bins is not None:
            from .cut import cut

            # cut
            try:
                inp = cut(inp, self.bins, include_lowest=True)
            except TypeError:  # pragma: no cover
                raise TypeError("bins argument only works with numeric data.")

            self.bins = None
            self.convert_index_to_interval = True
            return self.new_series(
                [inp],
                shape=(np.nan,),
                index_value=parse_index(pd.CategoricalIndex([]), inp, store_data=False),
                name=inp.name,
                dtype=test_series.dtype,
            )
        else:
            return self.new_series(
                [inp],
                shape=(np.nan,),
                index_value=parse_index(test_series.index, store_data=False),
                name=inp.name,
                dtype=test_series.dtype,
            )

    @classmethod
    def tile(cls, op: "DataFrameValueCounts"):
        inp = op.input
        out = op.outputs[0]

        if len(inp.chunks) == 1:
            chunk_op = op.copy().reset_key()
            chunk_param = out.params
            chunk_param["index"] = (0,)
            chunk = chunk_op.new_chunk(inp.chunks, kws=[chunk_param])

            new_op = op.copy()
            param = out.params
            param["chunks"] = [chunk]
            param["nsplits"] = ((np.nan,),)
            return new_op.new_seriess(op.inputs, kws=[param])

        inp = Series(inp)

        if op.dropna:
            inp = inp.dropna()

        inp = inp.groupby(inp, sort=not _keep_original_order).count(method=op.method)

        if op.normalize:
            if op.convert_index_to_interval:
                if has_unknown_shape(op.input):
                    yield
                inp = inp.truediv(op.input.shape[0], axis=0)
            else:
                inp = inp.truediv(inp.sum(), axis=0)

        if op.sort:
            inp = inp.sort_values(
                ascending=op.ascending,
                kind="mergesort" if _keep_original_order else "quicksort",
            )

            if op.nrows:
                # set to sort_values
                inp.op.nrows = op.nrows
        elif op.nrows:
            inp = inp.iloc[: op.nrows]

        ret = yield from recursive_tile(inp)

        chunks = []
        for c in ret.chunks:
            chunk_op = DataFrameValueCounts(
                convert_index_to_interval=op.convert_index_to_interval,
                stage=OperandStage.map,
            )
            chunk_params = c.params
            if op.convert_index_to_interval:
                # convert index to IntervalDtype
                chunk_params["index_value"] = parse_index(
                    pd.IntervalIndex([]), c, store_data=False
                )
            chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))

        new_op = op.copy()
        params = out.params
        params["chunks"] = chunks
        params["nsplits"] = ret.nsplits
        return new_op.new_seriess(out.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op: "DataFrameValueCounts"):
        if op.stage != OperandStage.map:
            in_data = ctx[op.input.key]
            if op.convert_index_to_interval:
                result = in_data.value_counts(
                    normalize=False,
                    sort=op.sort,
                    ascending=op.ascending,
                    bins=op.bins,
                    dropna=op.dropna,
                )
                if op.normalize:
                    result /= in_data.shape[0]
            else:
                try:
                    result = in_data.value_counts(
                        normalize=op.normalize,
                        sort=op.sort,
                        ascending=op.ascending,
                        bins=op.bins,
                        dropna=op.dropna,
                    )
                except ValueError:
                    in_data = in_data.copy()
                    result = in_data.value_counts(
                        normalize=op.normalize,
                        sort=op.sort,
                        ascending=op.ascending,
                        bins=op.bins,
                        dropna=op.dropna,
                    )
        else:
            result = ctx[op.input.key]
            # set index name to None to keep consistency with pandas
            result.index.name = None
        if op.convert_index_to_interval:
            # convert CategoricalDtype which generated in `cut`
            # to IntervalDtype
            result.index = result.index.astype("interval")
        if op.nrows:
            result = result.head(op.nrows)
        ctx[op.outputs[0].key] = result


def value_counts(
    series,
    normalize=False,
    sort=True,
    ascending=False,
    bins=None,
    dropna=True,
    method="auto",
):
    """
    Return a Series containing counts of unique values.

    The resulting object will be in descending order so that the
    first element is the most frequently-occurring element.
    Excludes NA values by default.

    Parameters
    ----------
    normalize : bool, default False
        If True then the object returned will contain the relative
        frequencies of the unique values.
    sort : bool, default True
        Sort by frequencies.
    ascending : bool, default False
        Sort in ascending order.
    bins : int, optional
        Rather than count values, group them into half-open bins,
        a convenience for ``pd.cut``, only works with numeric data.
    dropna : bool, default True
        Don't include counts of NaN.
    method : str, default 'auto'
        'auto', 'shuffle', or 'tree', 'tree' method provide
        a better performance, while 'shuffle' is recommended
        if aggregated result is very large, 'auto' will use
        'shuffle' method in distributed mode and use 'tree'
        in local mode.

    Returns
    -------
    Series

    See Also
    --------
    Series.count: Number of non-NA elements in a Series.
    DataFrame.count: Number of non-NA elements in a DataFrame.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> import mars.tensor as mt

    >>> s = md.Series([3, 1, 2, 3, 4, mt.nan])
    >>> s.value_counts().execute()
    3.0    2
    4.0    1
    2.0    1
    1.0    1
    dtype: int64

    With `normalize` set to `True`, returns the relative frequency by
    dividing all values by the sum of values.

    >>> s = md.Series([3, 1, 2, 3, 4, mt.nan])
    >>> s.value_counts(normalize=True).execute()
    3.0    0.4
    4.0    0.2
    2.0    0.2
    1.0    0.2
    dtype: float64

    **bins**

    Bins can be useful for going from a continuous variable to a
    categorical variable; instead of counting unique
    apparitions of values, divide the index in the specified
    number of half-open bins.

    >>> s.value_counts(bins=3).execute()
    (2.0, 3.0]      2
    (0.996, 2.0]    2
    (3.0, 4.0]      1
    dtype: int64

    **dropna**

    With `dropna` set to `False` we can also see NaN index values.

    >>> s.value_counts(dropna=False).execute()
    3.0    2
    NaN    1
    4.0    1
    2.0    1
    1.0    1
    dtype: int64
    """
    op = DataFrameValueCounts(
        normalize=normalize,
        sort=sort,
        ascending=ascending,
        bins=bins,
        dropna=dropna,
        method=method,
    )
    return op(series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import io
import os
import sys
import textwrap
import weakref
from typing import List, Callable, Type

from ..typing import OperandType, TileableType, SessionType
from .context import Context


class _LogWrapper:
    def __init__(self, ctx: Context, op: OperandType, log_path: str):
        self.ctx = ctx
        self.op = op
        self.log_path = log_path

        self.file = open(log_path, "w")
        self.stdout = sys.stdout

        self.raw_stdout = self.stdout
        while isinstance(self.raw_stdout, _LogWrapper):
            self.raw_stdout = self.raw_stdout.stdout

        # flag about registering log path
        self.is_log_path_registered = False

    def __enter__(self):
        self.file.__enter__()
        # set stdout
        sys.stdout = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.__exit__(exc_type, exc_val, exc_tb)
        # set back stdout
        sys.stdout = self.stdout

    def _register_log_path(self):
        if self.is_log_path_registered:
            return

        # register log path
        session_id = self.ctx.session_id
        tileable_op_key = self.op.tileable_op_key
        chunk_op_key = self.op.key
        worker_addr = self.ctx.local_address
        log_path = self.log_path

        self.ctx.register_custom_log_path(
            session_id, tileable_op_key, chunk_op_key, worker_addr, log_path
        )

        self.is_log_path_registered = True

    def write(self, data):
        self._register_log_path()

        # write into file
        self.file.write(data)
        # force flush to make sure `fetch_log` can get stdout in time
        self.file.flush()
        # write into previous stdout
        self.raw_stdout.write(data)

    def flush(self):
        self.raw_stdout.flush()


def redirect_custom_log(func: Callable[[Type, Context, OperandType], None]):
    """
    Redirect stdout to a file by wrapping ``Operand.execute(ctx, op)``
    """

    @functools.wraps(func)
    def wrap(cls, ctx: Context, op: OperandType):
        custom_log_dir = ctx.new_custom_log_dir()

        if custom_log_dir is None:
            return func(cls, ctx, op)

        log_path = os.path.join(custom_log_dir, op.key)

        with _LogWrapper(ctx, op, log_path):
            return func(cls, ctx, op)

    return wrap


_tileable_to_log_fetcher = weakref.WeakKeyDictionary()


class LogFetcher:
    def __init__(self, tileable_op_key: str, session: SessionType):
        self._tileable_op_key = tileable_op_key
        self._session = session
        self._chunk_op_key_to_result = dict()
        self._chunk_op_key_to_offsets = dict()

    def __len__(self):
        return len(self._chunk_op_key_to_result)

    @property
    def chunk_op_keys(self) -> List[str]:
        return list(self._chunk_op_key_to_result.keys())

    @property
    def results(self) -> list:
        return list(self._chunk_op_key_to_result.values())

    @property
    def offsets(self) -> List[List[int]]:
        return list(self._chunk_op_key_to_offsets.values())

    def fetch(self, offsets: List[int] = None, sizes: List[int] = None):
        if offsets is None:
            offsets = self._chunk_op_key_to_offsets

        if sizes is None:
            sizes = 1 * 1024**2  # 1M each time

        result: dict = self._session.fetch_tileable_op_logs(
            self._tileable_op_key, offsets=offsets, sizes=sizes
        )

        if result is None:
            return

        for chunk_key, chunk_result in result.items():
            self._chunk_op_key_to_result[chunk_key] = chunk_result["log"]
            self._chunk_op_key_to_offsets[chunk_key] = chunk_result["offset"]

    def _display(self, representation: bool = True):
        if len(self) == 1:
            content = next(iter(self._chunk_op_key_to_result.values()))
            return repr(content) if representation else str(content)

        sio = io.StringIO()
        for chunk_op_key, content in self._chunk_op_key_to_result.items():
            sio.write(
                textwrap.dedent(
                    f"""
                Chunk op key: {chunk_op_key}
                Out:
                {content}"""
                )
            )
        result = sio.getvalue()
        return repr(result) if representation else str(result)

    def __repr__(self):
        return self._display(True)

    def __str__(self):
        return self._display(False)


def fetch(
    tileables: List[TileableType],
    session: SessionType,
    offsets: List[int] = None,
    sizes: List[int] = None,
):
    log_fetchers = []
    for tileable in tileables:
        tileable = tileable.data if hasattr(tileable, "data") else tileable

        if tileable not in _tileable_to_log_fetcher:
            _tileable_to_log_fetcher[tileable] = LogFetcher(tileable.op.key, session)

        log_fetcher = _tileable_to_log_fetcher[tileable]
        log_fetcher.fetch(offsets=offsets, sizes=sizes)
        log_fetchers.append(log_fetcher)
    return log_fetchers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import inject_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand
class TensorFloatPower(TensorBinOp):
    _op_type_ = OperandDef.FLOAT_POWER
    _func_name = "float_power"

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, "issparse") and x1.issparse():
            return True
        return False


@inject_dtype(np.float64)
def float_power(x1, x2, out=None, where=None, **kwargs):
    """
    First tensor elements raised to powers from second array, element-wise.

    Raise each base in `x1` to the positionally-corresponding power in `x2`.
    `x1` and `x2` must be broadcastable to the same shape. This differs from
    the power function in that integers, float16, and float32  are promoted to
    floats with a minimum precision of float64 so that the result is always
    inexact.  The intent is that the function will return a usable result for
    negative powers and seldom overflow for positive powers.

    Parameters
    ----------
    x1 : array_like
        The bases.
    x2 : array_like
        The exponents.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        The bases in `x1` raised to the exponents in `x2`.

    See Also
    --------
    power : power function that preserves type

    Examples
    --------
    Cube each element in a list.

    >>> import mars.tensor as mt

    >>> x1 = range(6)
    >>> x1
    [0, 1, 2, 3, 4, 5]
    >>> mt.float_power(x1, 3).execute()
    array([   0.,    1.,    8.,   27.,   64.,  125.])

    Raise the bases to different exponents.

    >>> x2 = [1.0, 2.0, 3.0, 3.0, 2.0, 1.0]
    >>> mt.float_power(x1, x2).execute()
    array([  0.,   1.,   8.,  27.,  16.,   5.])

    The effect of broadcasting.

    >>> x2 = mt.array([[1, 2, 3, 3, 2, 1], [1, 2, 3, 3, 2, 1]])
    >>> x2.execute()
    array([[1, 2, 3, 3, 2, 1],
           [1, 2, 3, 3, 2, 1]])
    >>> mt.float_power(x1, x2).execute()
    array([[  0.,   1.,   8.,  27.,  16.,   5.],
           [  0.,   1.,   8.,  27.,  16.,   5.]])
    """
    op = TensorFloatPower(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging
import os
import time
from typing import Union, Dict, List, Optional, AsyncGenerator

from ... import oscar as mo
from ...core.entrypoints import init_extension_entrypoints
from ...metrics import init_metrics
from ...oscar.backends.ray.driver import RayActorDriver
from ...oscar.backends.ray.utils import (
    process_placement_to_address,
    node_placement_to_address,
    process_address_to_placement,
)
from ...oscar.backends.ray.pool import RayPoolState
from ...oscar.backends.router import Router
from ...oscar.errors import ReconstructWorkerError
from ...resource import Resource
from ...services.cluster.backends.base import (
    register_cluster_backend,
    AbstractClusterBackend,
)
from ...services import NodeRole
from ...utils import lazy_import, retry_callable
from ..utils import (
    load_config,
    get_third_party_modules_from_config,
)
from .service import start_supervisor, start_worker, stop_supervisor, stop_worker
from .session import (
    _new_session,
    new_session,
    AbstractSession,
    ensure_isolation_created,
)
from .pool import create_supervisor_actor_pool, create_worker_actor_pool

ray = lazy_import("ray")
logger = logging.getLogger(__name__)

# The default config file.
DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "rayconfig.yml"
)
# The default value for supervisor standalone (not share node with worker).
DEFAULT_SUPERVISOR_STANDALONE = False
# The default value for supervisor sub pool count.
DEFAULT_SUPERVISOR_SUB_POOL_NUM = 0


def _load_config(config: Union[str, Dict] = None):
    return load_config(config, default_config_file=DEFAULT_CONFIG_FILE)


@register_cluster_backend
class RayClusterBackend(AbstractClusterBackend):
    name = "ray"

    def __init__(self, lookup_address: str, cluster_state_ref):
        self._supervisors = [n.strip() for n in lookup_address.split(",")]
        self._cluster_state_ref = cluster_state_ref

    @classmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: str, pool_address: str
    ) -> "RayClusterBackend":
        try:
            ref = await mo.create_actor(
                ClusterStateActor,
                uid=ClusterStateActor.default_uid(),
                address=lookup_address,
            )
        except mo.ActorAlreadyExist:  # pragma: no cover
            ref = await mo.actor_ref(
                ClusterStateActor.default_uid(), address=lookup_address
            )
        return cls(lookup_address, ref)

    async def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        yield self._supervisors

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        return self._supervisors

    async def new_worker(self, worker_address):
        return await self._cluster_state_ref.new_worker(worker_address)

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        return await self._cluster_state_ref.request_worker(
            worker_cpu, worker_mem, timeout
        )

    async def release_worker(self, address: str):
        return await self._cluster_state_ref.release_worker(address)

    async def reconstruct_worker(self, address: str):
        return await self._cluster_state_ref.reconstruct_worker(address)

    def get_cluster_state_ref(self):
        return self._cluster_state_ref


class ClusterStateActor(mo.StatelessActor):
    def __init__(self):
        self._worker_cpu, self._worker_mem, self._config = None, None, None
        self._pg_name, self._band_to_resource, self._worker_modules = None, None, None
        self._pg_counter = itertools.count()
        self._worker_count = 0
        self._workers = {}
        self._releasing_tasks = {}
        self._reconstructing_tasks = {}

    async def __post_create__(self):
        self._pg_name, _, _ = process_address_to_placement(self.address)

    def set_config(self, worker_cpu, worker_mem, config):
        self._worker_cpu, self._worker_mem, self._config = (
            worker_cpu,
            worker_mem,
            config,
        )
        # TODO(chaokunyang) Support gpu
        self._band_to_resource = {
            "numa-0": Resource(num_cpus=self._worker_cpu, mem_bytes=self._worker_mem)
        }
        self._worker_modules = get_third_party_modules_from_config(
            self._config, NodeRole.WORKER
        )

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> Optional[str]:
        worker_cpu = worker_cpu or self._worker_cpu
        worker_mem = worker_mem or self._worker_mem
        bundle = {
            "CPU": worker_cpu,
            # "memory": worker_mem or self._worker_mem
        }
        band_to_resource = {
            "numa-0": Resource(num_cpus=worker_cpu, mem_bytes=worker_mem)
        }
        start_time = time.time()
        logger.info("Start to request worker with resource %s.", bundle)
        # TODO rescale ray placement group instead of creating new placement group
        pg_name = f"{self._pg_name}_{next(self._pg_counter)}"
        pg = ray.util.placement_group(name=pg_name, bundles=[bundle], strategy="SPREAD")
        create_pg_timeout = timeout or 120
        try:
            await asyncio.wait_for(pg.ready(), timeout=create_pg_timeout)
        except asyncio.CancelledError:  # pragma: no cover
            logger.warning(
                "Request worker with placement group %s in %s seconds canceled.",
                pg.bundle_specs,
                create_pg_timeout,
            )
            ray.util.remove_placement_group(pg)
            return None
        except asyncio.TimeoutError:
            logger.warning(
                "Request worker failed, "
                "can not create placement group %s in %s seconds.",
                pg.bundle_specs,
                create_pg_timeout,
            )
            ray.util.remove_placement_group(pg)
            return None
        logger.info(
            "Creating placement group %s took %.4f seconds",
            pg.bundle_specs,
            time.time() - start_time,
        )
        worker_address = process_placement_to_address(pg_name, 0, 0)
        worker_pool = await self.create_worker(worker_address)
        await self.start_worker(worker_address, band_to_resource=band_to_resource)
        logger.info(
            "Request worker %s succeeds in %.4f seconds",
            worker_address,
            time.time() - start_time,
        )
        self._workers[worker_address] = (worker_pool, pg)
        return worker_address

    async def create_worker(self, worker_address):
        start_time = time.time()
        worker_pool = await create_worker_actor_pool(
            worker_address,
            self._band_to_resource,
            modules=self._worker_modules,
            metrics=self._config.get("metrics", {}),
        )
        logger.info(
            "Create worker node %s succeeds in %.4f seconds.",
            worker_address,
            time.time() - start_time,
        )
        return worker_pool

    async def start_worker(self, worker_address, band_to_resource=None):
        self._worker_count += 1
        start_time = time.time()
        band_to_resource = band_to_resource or self._band_to_resource
        await start_worker(
            worker_address, self.address, band_to_resource, config=self._config
        )
        worker_pool = ray.get_actor(worker_address)
        await worker_pool.mark_service_ready.remote()
        logger.info(
            "Start services on worker %s succeeds in %.4f seconds.",
            worker_address,
            time.time() - start_time,
        )
        return worker_pool

    async def release_worker(self, address: str):
        logger.info("Start to release worker %s", address)
        task = self._reconstructing_tasks.get(address)
        if task is not None:
            task.cancel()

        task = self._releasing_tasks.get(address)
        if task is not None:
            logger.info("Waiting for releasing worker %s", address)
            return await task

        async def _release_worker():
            await stop_worker(address, self._config)
            pool, pg = self._workers.pop(address)
            await pool.actor_pool.remote("stop")
            if "COV_CORE_SOURCE" in os.environ:  # pragma: no cover
                try:
                    # must clean up first, or coverage info lost
                    await pool.cleanup.remote()
                except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                    pass
            ray.kill(pool.main_pool)
            ray.util.remove_placement_group(pg)
            logger.info("Released worker %s", address)

        task = asyncio.create_task(_release_worker())
        task.add_done_callback(lambda _: self._releasing_tasks.pop(address, None))
        self._releasing_tasks[address] = task
        return await task

    async def reconstruct_worker(self, address: str):
        task = self._releasing_tasks.get(address)
        if task is not None:
            raise ReconstructWorkerError(
                f"Can't reconstruct releasing worker {address}"
            )

        task = self._reconstructing_tasks.get(address)
        if task is not None:
            logger.info("Waiting for reconstruct worker %s", address)
            return await task

        async def _reconstruct_worker():
            logger.info("Reconstruct worker %s", address)
            actor = ray.get_actor(address)
            # ray call will error when actor is restarting
            state = await retry_callable(
                actor.state.remote, ex_type=ray.exceptions.RayActorError, sync=False
            )()
            if state == RayPoolState.SERVICE_READY:
                logger.info("Worker %s is service ready.")
                return

            if state == RayPoolState.INIT:
                await actor.start.remote()
            else:
                assert state == RayPoolState.POOL_READY

            start_time = time.time()
            await start_worker(
                address, self.address, self._band_to_resource, config=self._config
            )
            await actor.mark_service_ready.remote()
            logger.info(
                "Start services on worker %s succeeds in %.4f seconds.",
                address,
                time.time() - start_time,
            )

        task = asyncio.create_task(_reconstruct_worker())
        task.add_done_callback(lambda _: self._reconstructing_tasks.pop(address, None))
        self._reconstructing_tasks[address] = task
        return await task


async def new_cluster(
    cluster_name: str = None,
    supervisor_mem: int = 1 * 1024**3,
    worker_num: int = 1,
    worker_cpu: int = 2,
    worker_mem: int = 2 * 1024**3,
    config: Union[str, Dict] = None,
    **kwargs,
):
    cluster_name = cluster_name or f"ray-cluster-{int(time.time())}"
    if not ray.is_initialized():
        logger.warning("Ray is not started, start the local ray cluster by `ray.init`.")
        # add 16 logical cpus for other computing in ray.
        ray.init(num_cpus=16 + worker_num * worker_cpu)
    ensure_isolation_created(kwargs)
    if kwargs:  # pragma: no cover
        raise TypeError(f"new_cluster got unexpected arguments: {list(kwargs)}")
    n_supervisor_process = kwargs.get(
        "n_supervisor_process", DEFAULT_SUPERVISOR_SUB_POOL_NUM
    )
    cluster = RayCluster(
        cluster_name,
        supervisor_mem,
        worker_num,
        worker_cpu,
        worker_mem,
        config,
        n_supervisor_process=n_supervisor_process,
    )
    try:
        await cluster.start()
        return await RayClient.create(cluster)
    except Exception as ex:  # pragma: no cover
        # cleanup the cluster if failed.
        try:
            await cluster.stop()
        except Exception as stop_ex:
            raise stop_ex from ex
        raise ex


def new_cluster_in_ray(**kwargs):
    isolation = ensure_isolation_created(kwargs)
    coro = new_cluster(**kwargs)
    fut = asyncio.run_coroutine_threadsafe(coro, isolation.loop)
    client = fut.result()
    client.session.as_default()
    return client


new_cluster_in_ray.__doc__ = new_cluster.__doc__


def new_ray_session(
    address: str = None,
    session_id: str = None,
    default: bool = True,
    **new_cluster_kwargs,
) -> AbstractSession:
    """

    Parameters
    ----------
    address: str
        mars web server address.
    session_id: str
        session id. If not specified, will be generated automatically.
    default: bool
        whether set the session as default session.
    new_cluster_kwargs:
        See `new_cluster` arguments.
    """
    client = None
    if not address:
        client = new_cluster_in_ray(**new_cluster_kwargs)
        session_id = session_id or client.session.session_id
        address = client.address
    session = new_session(
        address=address, session_id=session_id, backend="mars", default=default
    )
    session._ray_client = client
    if default:
        # SyncSession set isolated_session as default session instead.
        AbstractSession.default._ray_client = client
    return session


class RayCluster:
    _supervisor_pool: "ray.actor.ActorHandle"
    _worker_pools: List["ray.actor.ActorHandle"]

    def __init__(
        self,
        cluster_name: str,
        supervisor_mem: int = 1 * 1024**3,
        worker_num: int = 1,
        worker_cpu: int = 2,
        worker_mem: int = 4 * 1024**3,
        config: Union[str, Dict] = None,
        n_supervisor_process: int = DEFAULT_SUPERVISOR_SUB_POOL_NUM,
    ):
        # load third party extensions.
        init_extension_entrypoints()
        self._cluster_name = cluster_name
        self._supervisor_mem = supervisor_mem
        self._n_supervisor_process = n_supervisor_process
        self._worker_num = worker_num
        self._worker_cpu = worker_cpu
        self._worker_mem = worker_mem
        # load config file to dict.
        self._config = load_config(config, default_config_file=DEFAULT_CONFIG_FILE)
        self.supervisor_address = None
        # Hold actor handles to avoid being freed
        self._supervisor_pool = None
        self._worker_addresses = []
        self._worker_pools = []
        self._stopped = False
        self._cluster_backend = None
        self.web_address = None

    async def start(self):
        logging.basicConfig(
            format=ray.ray_constants.LOGGER_FORMAT, level=logging.INFO, force=True
        )
        logger.info("Start cluster with config %s", self._config)
        # init metrics to guarantee metrics use in driver
        metric_configs = self._config.get("metrics", {})
        metric_backend = metric_configs.get("backend")
        init_metrics(metric_backend, config=metric_configs.get(metric_backend))
        address_to_resources = dict()
        supervisor_standalone = (
            self._config.get("cluster", {})
            .get("ray", {})
            .get("supervisor", {})
            .get("standalone", DEFAULT_SUPERVISOR_STANDALONE)
        )
        supervisor_sub_pool_num = (
            self._config.get("cluster", {})
            .get("ray", {})
            .get("supervisor", {})
            .get("sub_pool_num", self._n_supervisor_process)
        )
        from ...storage.ray import support_specify_owner

        if not support_specify_owner():  # pragma: no cover
            logger.warning(
                "Current installed ray version does not support specify owner, "
                "autoscale may not work."
            )
            # config['scheduling']['autoscale']['enabled'] = False
        self.supervisor_address = process_placement_to_address(self._cluster_name, 0, 0)
        if "cluster" not in self._config:  # pragma: no cover
            self._config["cluster"] = dict()
        self._config["cluster"]["lookup_address"] = self.supervisor_address
        address_to_resources[node_placement_to_address(self._cluster_name, 0)] = {
            "CPU": 1,
            # "memory": self._supervisor_mem,
        }
        worker_addresses = []
        if supervisor_standalone:
            for worker_index in range(1, self._worker_num + 1):
                worker_address = process_placement_to_address(
                    self._cluster_name, worker_index, 0
                )
                worker_addresses.append(worker_address)
                worker_node_address = node_placement_to_address(
                    self._cluster_name, worker_index
                )
                address_to_resources[worker_node_address] = {
                    "CPU": self._worker_cpu,
                    # "memory": self._worker_mem,
                }
        else:
            for worker_index in range(self._worker_num):
                worker_process_index = (
                    supervisor_sub_pool_num + 1 if worker_index == 0 else 0
                )
                worker_address = process_placement_to_address(
                    self._cluster_name, worker_index, worker_process_index
                )
                worker_addresses.append(worker_address)
                worker_node_address = node_placement_to_address(
                    self._cluster_name, worker_index
                )
                address_to_resources[worker_node_address] = {
                    "CPU": self._worker_cpu,
                    # "memory": self._worker_mem,
                }
        mo.setup_cluster(address_to_resources)

        # third party modules from config
        supervisor_modules = get_third_party_modules_from_config(
            self._config, NodeRole.SUPERVISOR
        )

        # create supervisor actor pool
        supervisor_pool_coro = asyncio.create_task(
            create_supervisor_actor_pool(
                self.supervisor_address,
                n_process=supervisor_sub_pool_num,
                main_pool_cpus=0,
                sub_pool_cpus=0,
                modules=supervisor_modules,
                metrics=self._config.get("metrics", {}),
            )
        )
        worker_pools = [
            asyncio.create_task(
                create_worker_actor_pool(
                    addr,
                    {
                        "numa-0": Resource(
                            num_cpus=self._worker_cpu, mem_bytes=self._worker_mem
                        )
                    },
                    modules=get_third_party_modules_from_config(
                        self._config, NodeRole.WORKER
                    ),
                    metrics=self._config.get("metrics", {}),
                )
            )
            for addr in worker_addresses
        ]
        self._supervisor_pool = await supervisor_pool_coro
        logger.info("Create supervisor on node %s succeeds.", self.supervisor_address)
        self._cluster_backend = await RayClusterBackend.create(
            NodeRole.WORKER, self.supervisor_address, self.supervisor_address
        )
        cluster_state_ref = self._cluster_backend.get_cluster_state_ref()
        await self._cluster_backend.get_cluster_state_ref().set_config(
            self._worker_cpu, self._worker_mem, self._config
        )
        # start service
        await start_supervisor(self.supervisor_address, config=self._config)
        logger.info(
            "Start services on supervisor %s succeeds.", self.supervisor_address
        )
        await self._supervisor_pool.mark_service_ready.remote()
        worker_pools = await asyncio.gather(*worker_pools)
        logger.info("Create %s workers succeeds.", len(worker_pools))
        await asyncio.gather(
            *[cluster_state_ref.start_worker(addr) for addr in worker_addresses]
        )
        logger.info("Start services on %s workers succeeds.", len(worker_addresses))
        for worker_address, worker_pool in zip(worker_addresses, worker_pools):
            self._worker_addresses.append(worker_address)
            self._worker_pools.append(worker_pool)

        from ...services.web.supervisor import WebActor

        web_actor = await mo.actor_ref(
            WebActor.default_uid(), address=self.supervisor_address
        )
        self.web_address = await web_actor.get_web_address()
        logger.warning("Web service started at %s", self.web_address)

    async def stop(self):
        if not self._stopped:
            try:
                for worker_address in self._worker_addresses:
                    await stop_worker(worker_address, self._config)
                for pool in self._worker_pools:
                    await pool.actor_pool.remote("stop")
                if self._supervisor_pool is not None:
                    await stop_supervisor(self.supervisor_address, self._config)
                    await self._supervisor_pool.actor_pool.remote("stop")
            finally:
                AbstractSession.reset_default()
                RayActorDriver.stop_cluster()
                Router.set_instance(None)
            self._stopped = True


class RayClient:
    def __init__(self, cluster: RayCluster, session: AbstractSession):
        self._cluster = cluster
        self._address = cluster.supervisor_address
        self._session = session
        # hold ray cluster by client to avoid actor handle out-of-scope
        session._ray_client = self

    @classmethod
    async def create(cls, cluster: RayCluster) -> "RayClient":
        session = await _new_session(cluster.supervisor_address, default=True)
        client = RayClient(cluster, session)
        AbstractSession.default._ray_client = client
        return client

    @property
    def address(self):
        return self._session.address

    @property
    def session(self):
        return self._session

    @property
    def web_address(self):
        return self._cluster.web_address

    async def __aenter__(self):
        return self

    async def __aexit__(self, *_):
        await self._stop()

    def stop(self):
        isolation = ensure_isolation_created({})
        fut = asyncio.run_coroutine_threadsafe(self._stop(), isolation.loop)
        return fut.result()

    async def _stop(self):
        await self._cluster.stop()
        AbstractSession.reset_default()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorFFTMixin, validate_fft, TensorHermitianFFT


class TensorHFFT(TensorHermitianFFT, TensorFFTMixin):
    _op_type_ = OperandDef.HFFT

    def __init__(self, n=None, axis=-1, norm=None, **kw):
        super().__init__(_n=n, _axis=axis, _norm=norm, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.n is not None:
            new_shape[op.axis] = op.n
        else:
            new_shape[op.axis] = 2 * (shape[op.axis] - 1)
        return tuple(new_shape)


def hfft(a, n=None, axis=-1, norm=None):
    """
    Compute the FFT of a signal that has Hermitian symmetry, i.e., a real
    spectrum.

    Parameters
    ----------
    a : array_like
        The input tensor.
    n : int, optional
        Length of the transformed axis of the output. For `n` output
        points, ``n//2 + 1`` input points are necessary.  If the input is
        longer than this, it is cropped.  If it is shorter than this, it is
        padded with zeros.  If `n` is not given, it is determined from the
        length of the input along the axis specified by `axis`.
    axis : int, optional
        Axis over which to compute the FFT. If not given, the last
        axis is used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.
        The length of the transformed axis is `n`, or, if `n` is not given,
        ``2*m - 2`` where ``m`` is the length of the transformed axis of
        the input. To get an odd number of output points, `n` must be
        specified, for instance as ``2*m - 1`` in the typical case,

    Raises
    ------
    IndexError
        If `axis` is larger than the last axis of `a`.

    See also
    --------
    rfft : Compute the one-dimensional FFT for real input.
    ihfft : The inverse of `hfft`.

    Notes
    -----
    `hfft`/`ihfft` are a pair analogous to `rfft`/`irfft`, but for the
    opposite case: here the signal has Hermitian symmetry in the time
    domain and is real in the frequency domain. So here it's `hfft` for
    which you must supply the length of the result if it is to be odd.

    * even: ``ihfft(hfft(a, 2*len(a) - 2) == a``, within roundoff error,
    * odd: ``ihfft(hfft(a, 2*len(a) - 1) == a``, within roundoff error.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> signal = mt.array([1, 2, 3, 4, 3, 2])
    >>> mt.fft.fft(signal).execute()
    array([ 15.+0.j,  -4.+0.j,   0.+0.j,  -1.-0.j,   0.+0.j,  -4.+0.j])
    >>> mt.fft.hfft(signal[:4]).execute() # Input first half of signal
    array([ 15.,  -4.,   0.,  -1.,   0.,  -4.])
    >>> mt.fft.hfft(signal, 6).execute()  # Input entire signal and truncate
    array([ 15.,  -4.,   0.,  -1.,   0.,  -4.])


    >>> signal = mt.array([[1, 1.j], [-1.j, 2]])
    >>> (mt.conj(signal.T) - signal).execute()   # check Hermitian symmetry
    array([[ 0.-0.j,  0.+0.j],
           [ 0.+0.j,  0.-0.j]])
    >>> freq_spectrum = mt.fft.hfft(signal)
    >>> freq_spectrum.execute()
    array([[ 1.,  1.],
           [ 2., -2.]])

    """
    a = astensor(a)
    validate_fft(a, axis=axis, norm=norm)
    op = TensorHFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.float_))
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="binary_and")
class TensorHypot(TensorBinOp):
    _op_type_ = OperandDef.HYPOT
    _func_name = "hypot"


@infer_dtype(np.hypot)
def hypot(x1, x2, out=None, where=None, **kwargs):
    """
    Given the "legs" of a right triangle, return its hypotenuse.

    Equivalent to ``sqrt(x1**2 + x2**2)``, element-wise.  If `x1` or
    `x2` is scalar_like (i.e., unambiguously cast-able to a scalar type),
    it is broadcast for use with each element of the other argument.
    (See Examples)

    Parameters
    ----------
    x1, x2 : array_like
        Leg of the triangle(s).
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated array is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    z : Tensor
        The hypotenuse of the triangle(s).

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.hypot(3*mt.ones((3, 3)), 4*mt.ones((3, 3))).execute()
    array([[ 5.,  5.,  5.],
           [ 5.,  5.,  5.],
           [ 5.,  5.,  5.]])

    Example showing broadcast of scalar_like argument:

    >>> mt.hypot(3*mt.ones((3, 3)), [4]).execute()
    array([[ 5.,  5.,  5.],
           [ 5.,  5.,  5.],
           [ 5.,  5.,  5.]])
    """
    op = TensorHypot(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Generator

from ...mode import enter_mode
from ..entity import TileableGraph, ChunkGraph
from .base import AbstractGraphBuilder


class TileableGraphBuilder(AbstractGraphBuilder):
    _graph: TileableGraph

    def __init__(self, graph: TileableGraph):
        super().__init__(graph=graph)

    @enter_mode(build=True, kernel=True)
    def _build(self) -> Union[TileableGraph, ChunkGraph]:
        self._add_nodes(self._graph, list(self._graph.result_tileables), set())
        return self._graph

    def build(self) -> Generator[Union[TileableGraph, ChunkGraph], None, None]:
        yield self._build()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import recursive_tile, get_output_types
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import (
    KeyField,
    FunctionField,
    TupleField,
    DictField,
    BoolField,
)
from ...utils import enter_current_session, has_unknown_shape, quiet_stdio
from ..operands import DataFrameOperand, DataFrameOperandMixin, OutputType
from ..utils import (
    build_df,
    build_empty_df,
    build_series,
    parse_index,
    validate_output_types,
    build_empty_series,
)


class DataFrameMapChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.MAP_CHUNK

    _input = KeyField("input")
    _func = FunctionField("func")
    _args = TupleField("args")
    _kwargs = DictField("kwargs")
    _with_chunk_index = BoolField("with_chunk_index")

    def __init__(
        self,
        input=None,
        func=None,
        args=None,
        kwargs=None,
        output_types=None,
        with_chunk_index=None,
        **kw,
    ):
        super().__init__(
            _input=input,
            _func=func,
            _args=args,
            _kwargs=kwargs,
            _output_types=output_types,
            _with_chunk_index=with_chunk_index,
            **kw,
        )

    @property
    def input(self):
        return self._input

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def with_chunk_index(self):
        return self._with_chunk_index

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def _infer_attrs_by_call(self, df_or_series):
        test_obj = (
            build_df(df_or_series, size=2)
            if df_or_series.ndim == 2
            else build_series(df_or_series, size=2, name=df_or_series.name)
        )
        kwargs = self.kwargs or dict()
        if self.with_chunk_index:
            kwargs["chunk_index"] = (0,) * df_or_series.ndim
        with np.errstate(all="ignore"), quiet_stdio():
            obj = self._func(test_obj, *self._args, **kwargs)

        if obj.ndim == 2:
            output_type = OutputType.dataframe
            dtypes = obj.dtypes
            if obj.shape == test_obj.shape:
                shape = (df_or_series.shape[0], len(dtypes))
            else:  # pragma: no cover
                shape = (np.nan, len(dtypes))
        else:
            output_type = OutputType.series
            dtypes = pd.Series([obj.dtype], name=obj.name)
            if obj.shape == test_obj.shape:
                shape = df_or_series.shape
            else:
                shape = (np.nan,)

        index_value = parse_index(
            obj.index, df_or_series, self._func, self._args, self._kwargs
        )
        return {
            "output_type": output_type,
            "index_value": index_value,
            "shape": shape,
            "dtypes": dtypes,
        }

    def __call__(self, df_or_series, index=None, dtypes=None):
        output_type = (
            self.output_types[0]
            if self.output_types
            else get_output_types(df_or_series)[0]
        )
        shape = self._kwargs.pop("shape", None)

        if dtypes is not None:
            index = index if index is not None else pd.RangeIndex(-1)
            index_value = parse_index(
                index, df_or_series, self._func, self._args, self._kwargs
            )
            if shape is None:  # pragma: no branch
                shape = (
                    (np.nan,)
                    if output_type == OutputType.series
                    else (np.nan, len(dtypes))
                )
        else:
            # try run to infer meta
            try:
                attrs = self._infer_attrs_by_call(df_or_series)
                output_type = attrs["output_type"]
                index_value = attrs["index_value"]
                shape = attrs["shape"]
                dtypes = attrs["dtypes"]
            except:  # noqa: E722  # nosec
                if df_or_series.ndim == 1 or output_type == OutputType.series:
                    output_type = OutputType.series
                    index = index if index is not None else pd.RangeIndex(-1)
                    index_value = parse_index(
                        index, df_or_series, self._func, self._args, self._kwargs
                    )
                    dtypes = pd.Series([np.dtype(object)])
                    shape = (np.nan,)
                else:
                    raise TypeError(
                        "Cannot determine `output_type`, "
                        "you have to specify it as `dataframe` or `series`, "
                        "for dataframe, `dtypes` is required as well "
                        "if output_type='dataframe'"
                    )

        if output_type == OutputType.series:
            return self.new_series(
                [df_or_series],
                dtype=dtypes.iloc[0],
                shape=shape,
                index_value=index_value,
                name=dtypes.name,
            )
        else:
            # dataframe
            columns_value = parse_index(dtypes.index, store_data=True)
            return self.new_dataframe(
                [df_or_series],
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )

    @classmethod
    def tile(cls, op: "DataFrameMapChunk"):
        inp = op.input
        out = op.outputs[0]

        if inp.ndim == 2 and inp.chunk_shape[1] > 1:
            if has_unknown_shape(inp):
                yield
            # if input is a DataFrame, make sure 1 chunk on axis columns
            inp = yield from recursive_tile(inp.rechunk({1: inp.shape[1]}))

        out_chunks = []
        nsplits = [[]] if out.ndim == 1 else [[], [out.shape[1]]]
        pd_out_index = out.index_value.to_pandas()
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op.tileable_op_key = op.key
            if op.output_types[0] == OutputType.dataframe:
                if np.isnan(out.shape[0]):
                    shape = (np.nan, out.shape[1])
                else:
                    shape = (chunk.shape[0], out.shape[1])
                index_value = parse_index(pd_out_index, chunk, op.key)
                out_chunk = chunk_op.new_chunk(
                    [chunk],
                    shape=shape,
                    dtypes=out.dtypes,
                    index_value=index_value,
                    columns_value=out.columns_value,
                    index=(chunk.index[0], 0),
                )
                out_chunks.append(out_chunk)
                nsplits[0].append(out_chunk.shape[0])
            else:
                if np.isnan(out.shape[0]):
                    shape = (np.nan,)
                else:
                    shape = (chunk.shape[0],)
                index_value = parse_index(pd_out_index, chunk, op.key)
                out_chunk = chunk_op.new_chunk(
                    [chunk],
                    shape=shape,
                    index_value=index_value,
                    name=out.name,
                    dtype=out.dtype,
                    index=(chunk.index[0],),
                )
                out_chunks.append(out_chunk)
                nsplits[0].append(out_chunk.shape[0])

        params = out.params
        params["nsplits"] = tuple(tuple(ns) for ns in nsplits)
        params["chunks"] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "DataFrameMapChunk"):
        inp = ctx[op.input.key]
        out = op.outputs[0]
        if len(inp) == 0:
            if op.output_types[0] == OutputType.dataframe:
                ctx[out.key] = build_empty_df(out.dtypes)
            elif op.output_types[0] == OutputType.series:
                ctx[out.key] = build_empty_series(out.dtype, name=out.name)
            else:
                raise ValueError(f"Chunk can not be empty except for dataframe/series.")
            return

        kwargs = op.kwargs or dict()
        if op.with_chunk_index:
            kwargs["chunk_index"] = out.index
        ctx[out.key] = op.func(inp, *op.args, **kwargs)


def map_chunk(df_or_series, func, args=(), **kwargs):
    """
    Apply function to each chunk.

    Parameters
    ----------
    func : function
        Function to apply to each chunk.
    args : tuple
        Positional arguments to pass to func in addition to the array/series.
    **kwargs
        Additional keyword arguments to pass as keywords arguments to func.

    Returns
    -------
    Series or DataFrame
        Result of applying ``func`` to each chunk of the DataFrame or Series.

    See Also
    --------
    DataFrame.apply : Perform any type of operations.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame([[4, 9]] * 3, columns=['A', 'B'])
    >>> df.execute()
       A  B
    0  4  9
    1  4  9
    2  4  9

    Output type including Series or DataFrame will be auto inferred.

    >>> df.map_chunk(lambda c: c['A'] + c['B']).execute()
    0    13
    1    13
    2    13
    dtype: int64

    You can specify ``output_type`` by yourself if auto infer failed.

    >>> import pandas as pd
    >>> import numpy as np
    >>> df['c'] = ['s1', 's2', 's3']
    >>> df.map_chunk(lambda c: pd.concat([c['A'], c['c'].str.slice(1).astype(int)], axis=1)).execute()
    Traceback (most recent call last):
    TypeError: Cannot determine `output_type`, you have to specify it as `dataframe` or `series`...
    >>> df.map_chunk(lambda c: pd.concat([c['A'], c['c'].str.slice(1).astype(int)], axis=1),
    >>>              output_type='dataframe', dtypes=pd.Series([np.dtype(object), np.dtype(int)])).execute()
       A  c
    0  4  1
    1  4  2
    2  4  3
    """
    output_type = kwargs.pop("output_type", None)
    output_types = kwargs.pop("output_types", None)
    object_type = kwargs.pop("object_type", None)
    output_types = validate_output_types(
        output_type=output_type, output_types=output_types, object_type=object_type
    )
    output_type = output_types[0] if output_types else None
    if output_type:
        output_types = [output_type]
    index = kwargs.pop("index", None)
    dtypes = kwargs.pop("dtypes", None)
    with_chunk_index = kwargs.pop("with_chunk_index", False)

    op = DataFrameMapChunk(
        input=df_or_series,
        func=func,
        args=args,
        kwargs=kwargs,
        output_types=output_types,
        with_chunk_index=with_chunk_index,
    )
    return op(df_or_series, index=index, dtypes=dtypes)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from ...config import options
from ...core import OutputType, recursive_tile
from ...core.operand import OperandStage, MapReduceOperand
from ...serialization.serializables import AnyField, Int32Field, StringField, KeyField
from ...utils import ceildiv, has_unknown_shape, lazy_import
from ..initializer import DataFrame as asdataframe
from ..operands import DataFrameOperandMixin, DataFrameShuffleProxy

cudf = lazy_import("cudf")


class DuplicateOperand(MapReduceOperand, DataFrameOperandMixin):
    _input = KeyField("input")
    _subset = AnyField("subset")
    _keep = AnyField("keep")
    _method = StringField("method")

    # subset chunk, used for method 'subset_tree'
    _subset_chunk = KeyField("subset_chunk")
    # shuffle phase, used in shuffle method
    _shuffle_size = Int32Field("shuffle_size")

    @property
    def input(self):
        return self._input

    @property
    def subset(self):
        return self._subset

    @property
    def keep(self):
        return self._keep

    @property
    def method(self):
        return self._method

    @property
    def subset_chunk(self):
        return self._subset_chunk

    @property
    def shuffle_size(self):
        return self._shuffle_size

    @classmethod
    def _get_shape(cls, input_shape, op):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _gen_tileable_params(
        cls, op: "DuplicateOperand", input_params
    ):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _gen_chunk_params(cls, op: "DuplicateOperand", input_chunk):  # pragma: no cover
        raise NotImplementedError

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if self._subset_chunk is not None:
            self._subset_chunk = self._inputs[1]

    @classmethod
    def _tile_one_chunk(cls, op: "DuplicateOperand"):
        inp = op.input
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        chunk_op._method = None
        in_chunk = inp.chunks[0]
        chunk_params = cls._gen_chunk_params(chunk_op, in_chunk)
        chunk = chunk_op.new_chunk([in_chunk], kws=[chunk_params])

        params = out.params
        params["chunks"] = [chunk]
        params["nsplits"] = tuple((s,) for s in chunk.shape)
        new_op = op.copy()
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _get_map_output_types(cls, input_chunk, method: str):
        raise NotImplementedError

    @classmethod
    def _gen_map_chunks(cls, op: "DuplicateOperand", inp, method, **kw):
        chunks = inp.chunks
        map_chunks = []
        for c in chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._output_types = cls._get_map_output_types(c, method)
            chunk_op._method = method
            chunk_op.stage = OperandStage.map
            for k, v in kw.items():
                setattr(chunk_op, k, v)
            chunk_params = cls._gen_chunk_params(chunk_op, c)
            map_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))
        return map_chunks

    @classmethod
    def _gen_tree_chunks(cls, op: "DuplicateOperand", inp, method):
        from ..merge import DataFrameConcat

        out = op.outputs[0]
        combine_size = options.combine_size
        new_chunks = cls._gen_map_chunks(op, inp, method)
        while len(new_chunks) > 1:
            out_chunk_size = ceildiv(len(new_chunks), combine_size)
            out_chunks = []
            for i in range(out_chunk_size):
                in_chunks = new_chunks[i * combine_size : (i + 1) * combine_size]
                s = sum(c.shape[0] for c in in_chunks)
                if in_chunks[0].ndim == 2:
                    kw = dict(
                        dtypes=in_chunks[0].dtypes,
                        index_value=in_chunks[0].index_value,
                        columns_value=in_chunks[0].columns_value,
                        shape=(s, in_chunks[0].shape[1]),
                        index=(i, 0),
                    )
                else:
                    kw = dict(
                        dtype=in_chunks[0].dtype,
                        index_value=in_chunks[0].index_value,
                        name=in_chunks[0].name,
                        shape=(s,),
                        index=(i,),
                    )
                concat_chunk = DataFrameConcat(
                    output_types=in_chunks[0].op.output_types
                ).new_chunk(in_chunks, **kw)
                chunk_op = op.copy().reset_key()
                chunk_op._method = method
                chunk_op.stage = (
                    OperandStage.combine if out_chunk_size > 1 else OperandStage.agg
                )
                if out_chunk_size > 1 and method == "tree":
                    # for tree, chunks except last one should be dataframes,
                    chunk_op._output_types = (
                        [OutputType.dataframe]
                        if out_chunk_size > 1
                        else out.op.output_types
                    )
                elif method == "subset_tree":
                    # `subset_tree` will tile chunks that are always dataframes
                    chunk_op._output_types = [OutputType.dataframe]
                params = cls._gen_chunk_params(chunk_op, concat_chunk)
                if out.ndim == 1 and out_chunk_size == 1:
                    params["name"] = out.name
                out_chunks.append(chunk_op.new_chunk([concat_chunk], kws=[params]))
            new_chunks = out_chunks

        return new_chunks

    @classmethod
    def _tile_tree(cls, op: "DuplicateOperand", inp):
        out = op.outputs[0]

        params = out.params
        params["chunks"] = chunks = cls._gen_tree_chunks(op, inp, "tree")
        params["nsplits"] = tuple((s,) for s in chunks[0].shape)
        new_op = op.copy()
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _tile_subset_tree(cls, op: "DuplicateOperand", inp):
        # subset is available for DataFrame only
        inp = asdataframe(inp)
        out = op.outputs[0]
        subset = op.subset
        if subset is None:
            subset = inp.dtypes.index.tolist()
        # select subset first
        subset_df = yield from recursive_tile(inp[subset])
        # tree aggregate subset
        subset_chunk = cls._gen_tree_chunks(op, subset_df, "subset_tree")[0]

        out_chunks = []
        for c in inp.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._method = "subset_tree"
            chunk_op._subset_chunk = subset_chunk
            chunk_params = cls._gen_chunk_params(chunk_op, c)
            out_chunks.append(chunk_op.new_chunk([c, subset_chunk], kws=[chunk_params]))

        new_op = op.copy()
        params = out.params
        params["chunks"] = out_chunks
        splits = tuple(c.shape[0] for c in out_chunks)
        if out.ndim == 2:
            params["nsplits"] = (splits, inp.nsplits[1])
        else:
            params["nsplits"] = (splits,)
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def _tile_shuffle(cls, op: "DuplicateOperand", inp):
        out = op.outputs[0]

        map_chunks = cls._gen_map_chunks(
            op, inp, "shuffle", _shuffle_size=inp.chunk_shape[0]
        )
        proxy_chunk = DataFrameShuffleProxy(
            output_types=map_chunks[0].op.output_types
        ).new_chunk(map_chunks, shape=())
        reduce_chunks = []
        for i in range(len(map_chunks)):
            reduce_op = op.copy().reset_key()
            reduce_op._method = "shuffle"
            reduce_op.stage = OperandStage.reduce
            reduce_op.reducer_phase = "drop_duplicates"
            reduce_op.n_reducers = len(map_chunks)
            reduce_op.reducer_ordinal = i
            reduce_op._shuffle_size = inp.chunk_shape[0]
            reduce_op._output_types = op.output_types
            reduce_chunk_params = map_chunks[0].params
            reduce_chunk_params["index"] = (i,) + reduce_chunk_params["index"][1:]
            reduce_chunks.append(
                reduce_op.new_chunk([proxy_chunk], kws=[reduce_chunk_params])
            )

        put_back_proxy_chunk = DataFrameShuffleProxy(
            output_types=map_chunks[0].op.output_types
        ).new_chunk(reduce_chunks, shape=())
        put_back_chunks = []
        for i in range(len(map_chunks)):
            put_back_op = op.copy().reset_key()
            put_back_op._method = "shuffle"
            put_back_op.stage = OperandStage.reduce
            put_back_op.reducer_phase = "put_back"
            put_back_op.reducer_index = (i,)
            put_back_op.n_reducers = len(map_chunks)
            put_back_op.reducer_ordinal = i
            if out.ndim == 2:
                put_back_chunk_params = map_chunks[i].params
            else:
                put_back_chunk_params = out.params.copy()
                map_chunk_params = map_chunks[i].params
                put_back_chunk_params["index_value"] = map_chunk_params["index_value"]
                put_back_chunk_params["index"] = map_chunk_params["index"][:1]
            if out.ndim == 1:
                put_back_chunk_params["index"] = (i,)
            else:
                put_back_chunk_params["index"] = (i,) + put_back_chunk_params["index"][
                    1:
                ]
            put_back_chunk_params["shape"] = cls._get_shape(
                map_chunks[i].op.input.shape, op
            )
            put_back_chunks.append(
                put_back_op.new_chunk(
                    [put_back_proxy_chunk], kws=[put_back_chunk_params]
                )
            )

        new_op = op.copy()
        params = out.params
        params["chunks"] = put_back_chunks
        split = tuple(c.shape[0] for c in put_back_chunks)
        if out.ndim == 2:
            params["nsplits"] = (split, inp.nsplits[1])
        else:
            params["nsplits"] = (split,)
        return new_op.new_tileables([inp], kws=[params])

    @classmethod
    def tile(cls, op: "DuplicateOperand"):
        inp = op.input

        if len(inp.chunks) == 1:
            # one chunk
            return cls._tile_one_chunk(op)

        if inp.ndim == 2 and inp.chunk_shape[1] > 1:
            if has_unknown_shape(inp):
                yield
            inp = yield from recursive_tile(inp.rechunk({1: inp.shape[1]}))

        default_tile = cls._tile_tree

        if op.method == "auto":
            # if method == 'auto', pick appropriate method
            if np.isnan(inp.shape[0]) or op.subset is None:
                # if any unknown shape exist,
                # choose merge method
                return default_tile(op, inp)

            # check subset data to see if it's small enough
            subset_dtypes = inp.dtypes[op.subset]
            memory_usage = 0.0
            for s_dtype in subset_dtypes:
                if s_dtype.kind == "O" or not hasattr(s_dtype, "itemsize"):
                    # object, just use default tile
                    return default_tile(op, inp)
                else:
                    memory_usage += s_dtype.itemsize * inp.shape[0]
            if memory_usage <= options.chunk_store_limit:
                # if subset is small enough, use method 'subset_tree'
                r = yield from cls._tile_subset_tree(op, inp)
                return r
            else:
                return default_tile(op, inp)
        elif op.method == "subset_tree":
            r = yield from cls._tile_subset_tree(op, inp)
            return r
        elif op.method == "tree":
            return cls._tile_tree(op, inp)
        else:
            assert op.method == "shuffle"
            ret = cls._tile_shuffle(op, inp)
            if inspect.isgenerator(ret):
                return (yield from ret)
            else:
                return ret

    @classmethod
    def _drop_duplicates(cls, inp, op, subset=None, keep=None, ignore_index=None):
        if ignore_index is None:
            ignore_index = op.ignore_index
        if subset is None:
            subset = op.subset
        if keep is None:
            keep = op.keep
        if inp.ndim == 2:
            try:
                return inp.drop_duplicates(
                    subset=subset, keep=keep, ignore_index=ignore_index
                )
            except TypeError:
                # no ignore_index for pandas < 1.0
                ret = inp.drop_duplicates(subset=subset, keep=keep)
                if ignore_index:
                    ret.reset_index(drop=True, inplace=True)
                return ret
        else:
            return inp.drop_duplicates(keep=keep)

    @classmethod
    def _get_xdf(cls, x):
        if cudf is None:
            return pd
        elif isinstance(x, (pd.Index, pd.Series, pd.DataFrame)):  # pragma: no cover
            return pd
        else:  # pragma: no cover
            return cudf

    @classmethod
    def _execute_subset_tree_map(cls, ctx, op):
        out = op.outputs[0]
        idx = out.index[0]
        inp = ctx[op.input.key]
        xdf = cls._get_xdf(inp)

        # index would be (chunk_index, i)
        index = xdf.MultiIndex.from_arrays(
            [np.full(inp.shape[0], idx), np.arange(inp.shape[0])],
            names=["_chunk_index_", "_i_"],
        )
        inp = inp.set_index(index)
        ctx[out.key] = cls._drop_duplicates(inp, op, ignore_index=False)

    @classmethod
    def _execute_subset_tree_combine(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = cls._drop_duplicates(inp, op, ignore_index=False)

    @classmethod
    def _execute_subset_tree_agg(cls, ctx, op):
        inp = ctx[op.input.key]
        ret = cls._drop_duplicates(inp, op, ignore_index=False)
        ret = ret.index.to_frame()
        ret.reset_index(drop=True, inplace=True)
        ctx[op.outputs[0].key] = ret


def validate_subset(df, subset):
    if subset is None:
        return subset
    if not is_list_like(subset):
        subset = [subset]
    else:
        subset = list(subset)

    for s in subset:
        if s not in df.dtypes:
            raise KeyError(pd.Index([s]))

    return subset
